#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import pandas as pd
import numpy as np
import numba

from typing import Callable, Tuple
from typing_extensions import Literal

from dios import DictOfSeries

from saqc.constants import *
from saqc.core.register import flagging
from saqc.lib.tools import customRoller, filterKwargs
from saqc.core import register, Flags


@flagging()
def flagChangePoints(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    stat_func: Callable[[np.ndarray, np.ndarray], float],
    thresh_func: Callable[[np.ndarray, np.ndarray], float],
    window: str | Tuple[str, str],
    min_periods: int | Tuple[int, int],
    closed: Literal["right", "left", "both", "neither"] = "both",
    reduce_window: str = None,
    reduce_func: Callable[[np.ndarray, np.ndarray], int] = lambda x, _: x.argmax(),
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Flag data points, where the parametrization of the process, the data is assumed to
    generate by, significantly changes.

    The change points detection is based on a sliding window search.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        The reference variable, the deviation from wich determines the flagging.

    flags : saqc.flags
        A flags object, holding flags and additional information related to `data`.

    stat_func : Callable[numpy.array, numpy.array]
         A function that assigns a value to every twin window. The backward-facing
         window content will be passed as the first array, the forward-facing window
         content as the second.

    thresh_func : Callable[numpy.array, numpy.array]
        A function that determines the value level, exceeding wich qualifies a
        timestamps func value as denoting a change-point.

    window : str, tuple of string
        Size of the rolling windows the calculation is performed in. If it is a single
        frequency offset, it applies for the backward- and the forward-facing window.

        If two offsets (as a tuple) is passed the first defines the size of the
        backward facing window, the second the size of the forward facing window.

    min_periods : int or tuple of int
        Minimum number of observations in a window required to perform the changepoint
        test. If it is a tuple of two int, the first refer to the backward-,
        the second to the forward-facing window.

    closed : {'right', 'left', 'both', 'neither'}, default 'both'
        Determines the closure of the sliding windows.

    reduce_window : {None, str}, default None
        The sliding window search method is not an exact CP search method and usually there wont be
        detected a single changepoint, but a "region" of change around a changepoint.
        If `reduce_window` is given, for every window of size `reduce_window`, there
        will be selected the value with index `reduce_func(x, y)` and the others will be dropped.
        If `reduce_window` is None, the reduction window size equals the
        twin window size, the changepoints have been detected with.

    reduce_func : Callable[[numpy.ndarray, numpy.ndarray], int], default lambda x, y: x.argmax()
        A function that must return an index value upon input of two arrays x and y.
        First input parameter will hold the result from the stat_func evaluation for every
        reduction window. Second input parameter holds the result from the thresh_func evaluation.
        The default reduction function just selects the value that maximizes the stat_func.

    flag : float, default BAD
        flag to set.

    Returns
    -------
    """
    return _assignChangePointCluster(
        data,
        field,
        flags,
        stat_func=stat_func,
        thresh_func=thresh_func,
        window=window,
        min_periods=min_periods,
        closed=closed,
        reduce_window=reduce_window,
        reduce_func=reduce_func,
        set_flags=True,
        model_by_resids=False,
        assign_cluster=False,
        flag=flag,
        **kwargs,
    )


@register(mask=["field"], demask=[], squeeze=[])
def assignChangePointCluster(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    stat_func: Callable[[np.array, np.array], float],
    thresh_func: Callable[[np.array, np.array], float],
    window: str | Tuple[str, str],
    min_periods: int | Tuple[int, int],
    closed: Literal["right", "left", "both", "neither"] = "both",
    reduce_window: str = None,
    reduce_func: Callable[[np.ndarray, np.ndarray], float] = lambda x, _: x.argmax(),
    model_by_resids: bool = False,
    **kwargs,
):
    """
    Assigns label to the data, aiming to reflect continuous regimes of the processes
    the data is assumed to be generated by. The regime change points detection is
    based on a sliding window search.

    Note, that the cluster labels will be stored to the `field` field of the input
    data, so that the data that is clustered gets overridden.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        The reference variable, the deviation from wich determines the flagging.

    flags : saqc.flags
        A flags object, holding flags and additional informations related to `data`.

    stat_func : Callable[[numpy.array, numpy.array], float]
        A function that assigns a value to every twin window. Left window content will
        be passed to first variable,
        right window content will be passed to the second.

    thresh_func : Callable[numpy.array, numpy.array], float]
        A function that determines the value level, exceeding wich qualifies a
        timestamps func func value as denoting a changepoint.

    window : str, tuple of string
        Size of the rolling windows the calculation is performed in. If it is a single
        frequency offset, it applies for the backward- and the forward-facing window.

        If two offsets (as a tuple) is passed the first defines the size of the
        backward facing window, the second the size of the forward facing window.

    min_periods : int or tuple of int
        Minimum number of observations in a window required to perform the changepoint
        test. If it is a tuple of two int, the first refer to the backward-,
        the second to the forward-facing window.

    closed : {'right', 'left', 'both', 'neither'}, default 'both'
        Determines the closure of the sliding windows.

    reduce_window : {None, str}, default None
        The sliding window search method is not an exact CP search method and usually
        there wont be detected a single changepoint, but a "region" of change around
        a changepoint. If `reduce_window` is given, for every window of size
        `reduce_window`, there will be selected the value with index `reduce_func(x,
        y)` and the others will be dropped. If `reduce_window` is None, the reduction
        window size equals the twin window size, the changepoints have been detected
        with.

    reduce_func : callable, default lambda x,y: x.argmax()
        A function that must return an index value upon input of two arrays x and y.
        First input parameter will hold the result from the stat_func evaluation for
        every reduction window. Second input parameter holds the result from the
        thresh_func evaluation. The default reduction function just selects the value
        that maximizes the stat_func.

    model_by_resids : bool, default False
        If True, the stat_funcs results are written, otherwise the regime labels.

    Returns
    -------
    """
    reserved = ["assign_cluster", "set_flags", "flag"]
    kwargs = filterKwargs(kwargs, reserved)
    return _assignChangePointCluster(
        data=data,
        field=field,
        flags=flags,
        stat_func=stat_func,
        thresh_func=thresh_func,
        window=window,
        min_periods=min_periods,
        closed=closed,
        reduce_window=reduce_window,
        reduce_func=reduce_func,
        model_by_resids=model_by_resids,
        **kwargs,
        # control args
        assign_cluster=True,
        set_flags=False,
    )


def _assignChangePointCluster(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    stat_func: Callable[[np.array, np.array], float],
    thresh_func: Callable[[np.array, np.array], float],
    window: str | Tuple[str, str],
    min_periods: int | Tuple[int, int],
    closed: Literal["right", "left", "both", "neither"] = "both",
    reduce_window: str = None,
    reduce_func: Callable[[np.ndarray, np.ndarray], float] = lambda x, _: x.argmax(),
    model_by_resids: bool = False,
    set_flags: bool = False,
    assign_cluster: bool = True,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    data_ser = data[field].dropna()
    if isinstance(window, (list, tuple)):
        bwd_window, fwd_window = window
    else:
        bwd_window = fwd_window = window

    if isinstance(window, (list, tuple)):
        bwd_min_periods, fwd_min_periods = min_periods
    else:
        bwd_min_periods = fwd_min_periods = min_periods

    if reduce_window is None:
        s = int(
            pd.Timedelta(bwd_window).total_seconds()
            + pd.Timedelta(fwd_window).total_seconds()
        )
        reduce_window = f"{s}s"

    roller = customRoller(data_ser, window=bwd_window, min_periods=bwd_min_periods)
    bwd_start, bwd_end = roller.window_indexer.get_window_bounds(
        len(data_ser), min_periods=bwd_min_periods, closed=closed
    )

    roller = customRoller(
        data_ser, window=fwd_window, forward=True, min_periods=fwd_min_periods
    )
    fwd_start, fwd_end = roller.window_indexer.get_window_bounds(
        len(data_ser), min_periods=fwd_min_periods, closed=closed
    )

    min_mask = ~(
        (fwd_end - fwd_start <= fwd_min_periods)
        | (bwd_end - bwd_start <= bwd_min_periods)
    )
    fwd_end = fwd_end[min_mask]
    split = bwd_end[min_mask]
    bwd_start = bwd_start[min_mask]
    masked_index = data_ser.index[min_mask]
    check_len = len(fwd_end)
    data_arr = data_ser.values

    try_to_jit = True
    jit_sf = numba.jit(stat_func, nopython=True)
    jit_tf = numba.jit(thresh_func, nopython=True)
    try:
        jit_sf(data_arr[bwd_start[0] : bwd_end[0]], data_arr[fwd_start[0] : fwd_end[0]])
        jit_tf(data_arr[bwd_start[0] : bwd_end[0]], data_arr[fwd_start[0] : fwd_end[0]])
        stat_func = jit_sf
        thresh_func = jit_tf
    except (numba.TypingError, numba.UnsupportedError, IndexError):
        try_to_jit = False

    args = data_arr, bwd_start, fwd_end, split, stat_func, thresh_func, check_len

    if try_to_jit:
        stat_arr, thresh_arr = _slidingWindowSearchNumba(*args)
    else:
        stat_arr, thresh_arr = _slidingWindowSearch(*args)

    result_arr = stat_arr > thresh_arr

    if model_by_resids:
        residues = pd.Series(np.nan, index=data[field].index)
        residues[masked_index] = stat_arr
        data[field] = residues
        flags[:, field] = UNFLAGGED
        return data, flags

    det_index = masked_index[result_arr]
    detected = pd.Series(True, index=det_index)
    if reduce_window:
        l = detected.shape[0]
        roller = customRoller(detected, window=reduce_window, min_periods=1)
        start, end = roller.window_indexer.get_window_bounds(
            num_values=l, min_periods=1, closed="both", center=True
        )

        detected = _reduceCPCluster(
            stat_arr[result_arr], thresh_arr[result_arr], start, end, reduce_func, l
        )
        det_index = det_index[detected]

    if assign_cluster:
        cluster = pd.Series(False, index=data[field].index)
        cluster[det_index] = True
        cluster = cluster.cumsum()
        # (better to start cluster labels with number one)
        cluster += 1
        data[field] = cluster
        flags[:, field] = UNFLAGGED

    if set_flags:
        flags[det_index, field] = flag
    return data, flags


@numba.jit(parallel=True, nopython=True)
def _slidingWindowSearchNumba(
    data_arr, bwd_start, fwd_end, split, stat_func, thresh_func, num_val
):
    stat_arr = np.zeros(num_val)
    thresh_arr = np.zeros(num_val)
    for win_i in numba.prange(0, num_val - 1):
        x = data_arr[bwd_start[win_i] : split[win_i]]
        y = data_arr[split[win_i] : fwd_end[win_i]]
        stat_arr[win_i] = stat_func(x, y)
        thresh_arr[win_i] = thresh_func(x, y)
    return stat_arr, thresh_arr


def _slidingWindowSearch(
    data_arr, bwd_start, fwd_end, split, stat_func, thresh_func, num_val
):
    stat_arr = np.zeros(num_val)
    thresh_arr = np.zeros(num_val)
    for win_i in range(0, num_val - 1):
        x = data_arr[bwd_start[win_i] : split[win_i]]
        y = data_arr[split[win_i] : fwd_end[win_i]]
        stat_arr[win_i] = stat_func(x, y)
        thresh_arr[win_i] = thresh_func(x, y)
    return stat_arr, thresh_arr


def _reduceCPCluster(stat_arr, thresh_arr, start, end, obj_func, num_val):
    out_arr = np.zeros(shape=num_val, dtype=bool)
    for win_i in numba.prange(0, num_val):
        s, e = start[win_i], end[win_i]
        x = stat_arr[s:e]
        y = thresh_arr[s:e]
        pos = s + obj_func(x, y) + 1
        out_arr[s:e] = False
        out_arr[pos] = True

    return out_arr



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Sequence, Union

import saqc
from saqc.constants import UNFLAGGED, BAD, FILTER_ALL
from saqc.lib.types import GenericFunction


class Generic:
    def processGeneric(
        self,
        field: str | Sequence[str],
        func: GenericFunction,
        target: str | Sequence[str] = None,
        flag: float = UNFLAGGED,
        dfilter: float = FILTER_ALL,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("processGeneric", locals())

    def flagGeneric(
        self,
        field: Union[str, Sequence[str]],
        func: GenericFunction,
        target: Union[str, Sequence[str]] = None,
        flag: float = BAD,
        dfilter: float = FILTER_ALL,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagGeneric", locals())



#!/usr/bin/env python
from __future__ import annotations

from copy import deepcopy, copy as shallowcopy
from typing import Mapping, Hashable, Any, Sequence, overload

from . import operators as ops
from . import pandas_bridge as pdextra
from . import lib

from .lib import (
    _CAST_POLICIES,
    _throw_MixedItype_err_or_warn,
    _find_least_common_itype,
)

from abc import abstractmethod
import pandas as pd
import operator as op
import functools as ftools

__author__ = "Bert Palm"
__email__ = "bert.palm@ufz.de"
__copyright__ = "Copyright 2018, Helmholtz-Zentrum für Umweltforschung GmbH - UFZ"


class _DiosBase:
    @property
    @abstractmethod
    def _constructor(self) -> type[_DiosBase]:
        raise NotImplementedError

    def _finalize(self, other: _DiosBase):
        self._attrs = other._attrs
        return self

    def __init__(
        self,
        data=None,
        columns=None,
        index=None,
        itype=None,
        cast_policy="save",
        fastpath=False,
    ):

        self._attrs = {}
        self.cast_policy = cast_policy  # set via property

        # we are called internally
        if fastpath:
            self._itype = itype or lib.ObjItype
            if data is not None:
                self._data = data
            else:
                # it is significantly faster, to provide an index and fill it,
                # than to successively build the index by adding data
                self._data = pd.Series(dtype="O", index=columns)

        else:

            if index is not None and not isinstance(index, pd.Index):
                index = pd.Index(index)

            # itype=None means infer the itype by the data, so we first set to the highest
            # possible itype, then insert data, then infer the best-fitting itype.
            if itype is None and index is None:
                self._itype = lib.ObjItype
            else:
                if index is not None:
                    self._itype = lib.get_itype(index)
                if itype is not None:
                    self._itype = lib.get_itype(itype)

            cols = pd.Index([] if columns is None else columns)
            if not cols.is_unique:
                raise ValueError("columns must be unique")
            self._data = pd.Series(dtype="O", index=cols)

            if data is not None:
                self._init_insert_data(data, columns, index)

        # self._data may still contain nans; at positions where
        # no data was present, but a column-name was given
        if self._data.hasnans:
            e = pd.Series(dtype="O", index=index)
            for c in self.columns[self._data.isna()]:
                self._insert(c, e.copy())

        self._data.index.name = "columns"

        # we try to infer the itype, but if we still have
        # no data, we will set the itype lazy, i.e. with
        # the first non-empty _insert()
        if itype is None:
            if self.empty:
                self._itype = "INFER"
            else:
                self._itype = _find_least_common_itype(self._data)
                if not self._itype.unique:
                    _throw_MixedItype_err_or_warn(self.itype)

    def _init_insert_data(self, data, columns, index):
        """Insert items of a iterable in self"""

        if pdextra.is_iterator(data):
            data = list(data)

        if _is_dios_like(data) and not data.columns.is_unique:
            raise ValueError("columns index must have unique values")

        if _is_dios_like(data) or isinstance(data, dict):
            if columns is None:
                pass  # data is dict-like
            else:
                data = {k: data[k] for k in data if k in columns}

        elif isinstance(data, pd.Series):
            name = data.name or 0
            if columns is not None and len(columns) > 0:
                name = self.columns[0]
            data = {name: data}

        elif pdextra.is_nested_list_like(data):
            if columns is None:
                data = {i: d for i, d in enumerate(data)}
            elif len(data) == len(columns):
                data = dict(zip(self.columns, data))
            else:
                raise ValueError(
                    f"{len(columns)} columns passed, data implies {len(data)} columns"
                )

        elif pdextra.is_list_like(data):
            name = 0 if columns is None or len(columns) < 1 else self.columns[0]
            data = {name: data}

        else:
            raise TypeError("data type not understood")

        for k in data:
            s = pd.Series(data[k], index=index, dtype=object).infer_objects()
            self._insert(k, s)

    # ----------------------------------------------------------------------
    # checks

    def _is_valid_columns_index(self, obj):
        if isinstance(obj, pd.Series) and obj.dtype == "O":
            return True
        return False

    # ----------------------------------------------------------------------
    # Indexing Methods

    def _insert(self, col, val):
        """Insert a fresh new value as pd.Series into self"""
        val = list(val) if pdextra.is_iterator(val) else val

        if _is_dios_like(val):
            val = val.squeeze()
            if not isinstance(val, pd.Series):
                raise ValueError(f"Cannot insert frame-like with more than one column")

        elif val is None:
            val = pd.Series()

        elif not isinstance(val, pd.Series):
            raise TypeError(
                f"Only data of type pandas.Series can be inserted, passed was {type(val)}"
            )

        # set the itype lazy, i.e. when first non-empty
        # column is inserted
        if self._itype == "INFER":
            if not val.empty:
                self._itype = lib.get_itype(val.index)
                # cast all pre-inserted empty series
                self._cast_all(self._itype, self.cast_policy)
                if not self._itype.unique:
                    _throw_MixedItype_err_or_warn(self._itype)
        else:
            val = lib.cast_to_itype(val, self.itype, policy=self.cast_policy)

        val.name = col
        self._data.at[col] = val.copy(deep=True)

    @overload
    def __getitem__(self, key: str | int) -> pd.Series:
        ...

    @overload
    def __getitem__(self, key: slice) -> pd.Series:
        ...

    @overload
    def __getitem__(self, key: "_DiosBase" | pd.DataFrame) -> "_DiosBase":
        ...

    @overload
    def __getitem__(self, key: Sequence[str | int]) -> "_DiosBase":
        ...

    def __getitem__(self, key):
        """dios[key] -> dios/series"""
        # scalar        -> select a column
        # slice         -> select rows (on all columns)
        # bool dios     -> select columns, select rows
        # mask          -> select rows (on all columns)
        # list-like     -> select columns

        if pdextra.is_scalar(key):
            # NOTE: we shallow copy, to prevent changes on the
            # index mirror back to us and may mess up the itype.
            s = self._data.at[key]
            s.index = s.index.copy()
            return s

        if isinstance(key, slice):
            return self._slice(key)

        if _is_dios_like(key):
            return self._getitem_bool_dios(key)

        if pdextra.is_bool_indexer(key):
            return self._getitem_bool_listlike(key)

        # select columns and let pandas handle it
        data = self._data.loc[key]
        if self._is_valid_columns_index(data):
            return self._constructor(
                data=data, itype=self.itype, cast_policy=self.cast_policy, fastpath=True
            )._finalize(self)

        raise TypeError(f"cannot index columns with this type, {type(key)}")

    def _slice(self, key):
        """slices self, return copy"""
        if key == slice(None):
            return self.copy()

        new = self.copy_empty(columns=True)
        for c, series in self.items():
            new._data.at[c] = series[key]
        return new

    def _getitem_bool_dios(self, key):
        """Select items by a boolean dios-like drop un-selected indices."""

        if not _is_bool_dios_like(key):
            raise ValueError("Must pass DictOfSeries with boolean values only")

        new = self.copy_empty(columns=True)
        for c, series in self.items():
            if c in key:
                val = key[c].reindex(index=series.index, fill_value=False)
                new._data.at[c] = series.loc[val]
        return new

    def _getitem_bool_listlike(self, key):
        new = self.copy_empty(columns=True)
        for c, series in self.items():
            new._data.at[c] = series.loc[key]
        return new

    def __setitem__(self, key, value):
        """dios[key] = value"""
        key = list(key) if pdextra.is_iterator(key) else key
        if isinstance(key, tuple):
            raise KeyError(f"{key}. tuples are not allowed")

        elif pdextra.is_hashable(key):
            if isinstance(value, pd.Series) or key not in self.columns:
                self._insert(key, value)
            elif _is_dios_like(value) or pdextra.is_nested_list_like(value):
                raise ValueError("Incompatible indexer with multi-dimensional value")
            else:
                self._data.at[key][:] = value

        else:
            data = self.__getitem__(key)
            assert isinstance(
                data, self.__class__
            ), f"getitem returned data of type {type(data)}"

            # special cases
            if _is_dios_like(value):
                self._setitem_dios(data, value)
            # NOTE: pd.Series also considered list-like
            elif pdextra.is_list_like(value):
                self._setitem_listlike(data, value)

            # default case
            else:
                for c, series in data.items():
                    series[:] = value
                    self._data.at[c][series.index] = series

    def _setitem_listlike(self, data, value):

        value = value.values if isinstance(value, pd.Series) else value

        if len(value) != len(data.columns):
            raise ValueError(
                f"array-like value of length {len(value)} could "
                f"not be broadcast to indexing result of shape "
                f"(.., {len(data.columns)})"
            )

        for i, (c, series) in enumerate(data.items()):
            series[:] = value[i]
            self._data.at[c][series.index] = series

    def _setitem_dios(self, data, value):
        """Write values from a dios-like to self.

        No justification or alignment of columns, but of indices.
        If value has missing indices, nan's are inserted at that
        locations, just like `series.loc[:]=val` or `df[:]=val` do.

        Eg.
         di[::2] = di[::3]   ->   di[::2]

            x |        x |            x |
        ===== |     ==== |       ====== |
        0   x |     0  z |       0    z |
        2   x |  =  3  z |   ->  2  NaN |
        4   x |     6  z |       4  NaN |
        6   x |                  6    z |

        Parameter
        ----------
        data : dios
            A maybe trimmed version of self
        value : dios, pd.Dataframe
            The value to set with the same column dimension like data
        """

        if len(data) != len(value.columns):
            raise ValueError(
                f"shape mismatch: values array of shape "
                f"(.., {len(value.columns)}) could not "
                f"be broadcast to indexing result of "
                f"shape (.., {len(data.columns)})"
            )

        for i, (c, series) in enumerate(data.items()):
            # .loc cannot handle empty series,
            # like `emptySeries.loc[:] = [1,2]`
            if series.empty:
                continue
            val = value[value.columns[i]]
            series.loc[:] = val
            self._data.at[c].loc[series.index] = series

    def __delitem__(self, key):
        del self._data[key]

    # ------------------------------------------------------------------------------
    # Base properties and basic dunder magic

    @property
    def attrs(self) -> dict[Hashable, Any]:
        """
        Dictionary of global attributes of this dataset.
        """
        if self._attrs is None:
            self._attrs = {}
        return self._attrs

    @attrs.setter
    def attrs(self, value: Mapping[Hashable, Any]) -> None:
        self._attrs = dict(value)

    @property
    def columns(self):
        """The column labels of the DictOfSeries"""
        return self._data.index

    @columns.setter
    def columns(self, cols):
        index = pd.Index(cols)
        if not index.is_unique:
            raise ValueError("columns index must have unique values")
        self._data.index = index
        # rename all columns
        for i, s in enumerate(self._data):
            s.name = index[i]

    @property
    def itype(self):
        """The ``Itype`` of the DictOfSeries.

        See :ref:`Itype documentation <doc_itype:Itype>` for more info.
        """
        if self._itype == "INFER":
            return None
        return self._itype

    @itype.setter
    def itype(self, itype):
        itype = lib.get_itype(itype)
        self._cast_all(itype, policy=self.cast_policy)
        self._itype = itype

    @property
    def cast_policy(self):
        """The policy to use for casting new columns if its initial itype does not fit.

        See :ref:`Itype documentation <doc_itype:Itype>` for more info.
        """
        return self._policy

    @cast_policy.setter
    def cast_policy(self, policy):
        if policy not in _CAST_POLICIES:
            raise ValueError(f"policy must be one of {_CAST_POLICIES}")
        self._policy = policy

    def _cast_all(self, itype, policy):
        c = "?"
        new = self.copy_empty()
        try:
            for c, series in self.items():
                new._data.at[c] = lib.cast_to_itype(series, itype, policy=policy)
        except Exception as e:
            raise type(e)(f"Column {c}: " + str(e)) from e

    def __len__(self):
        return len(self.columns)

    @property
    def empty(self):
        """Indicator whether DictOfSeries is empty.

        Returns
        -------
        bool :
            If DictOfSeries is empty, return True, if not return False.

        See Also
        --------
        DictOfSeries.dropempty : drop empty columns
        DictOfSeries.dropna : drop NAN's from a DictOfSeries
        pandas.Series.dropna : drop NAN's from a Series

        Notes
        -----
            If DictOfSeries contains only NaNs, it is still not considered empty. See the example below.

        Examples
        --------
        An example of an actual empty DictOfSeries.

        >>> di_empty = DictOfSeries(columns=['A'])
        >>> di_empty
        Empty DictOfSeries
        Columns: ['A']
        >>> di_empty.empty
        True

        If we only have NaNs in our DictOfSeries, it is not considered empty!
        We will need to drop the NaNs to make the DictOfSeries empty:

        >>> di = pd.DictOfSeries({'A' : [np.nan]})
        >>> di
            A |
        ===== |
        0 NaN |
        >>> di.empty
        False
        >>> di.dropna().empty
        True
        """
        return len(self) == 0 or all(s.empty for s in self._data)

    def __iter__(self):
        yield from self.columns

    def __reversed__(self):
        yield from reversed(self.columns)

    def __contains__(self, item):
        return item in self.columns

    def items(self):
        yield from self._data.items()

    # ----------------------------------------------------------------------
    # if copy.copy() is copy.copy(): return copy.copy().copy()

    def __deepcopy__(self, memo=None):
        return self.copy(deep=True)

    def __copy__(self):
        return self.copy(deep=False)

    def copy(self, deep=True):
        """Make a copy of this DictOfSeries' indices and data.

        Parameters
        ----------
        deep : bool, default True
            Make a deep copy, including a copy of the data and the indices.
            With deep=False neither the indices nor the data are copied.

        Returns
        -------
        copy : DictOfSeries

        See Also
        --------
        pandas.DataFrame.copy
        """
        data = self._data.copy()  # always copy the outer hull series
        if deep:
            for c, series in self.items():
                data.at[c] = series.copy()

        new = self._constructor(
            data=data, itype=self.itype, cast_policy=self.cast_policy, fastpath=True
        )

        copyfunc = deepcopy if deep else shallowcopy
        new._attrs = copyfunc(self._attrs)

        return new

    def copy_empty(self, columns=True):
        """
        Return a new DictOfSeries object, with same properties than the original.
        Parameters
        ----------
        columns: bool, default True
             If ``True``, the copy will have the same, but empty columns like the original.

        Returns
        -------
        DictOfSeries: empty copy

        Examples
        --------

        >>> di = DictOfSeries({'A': range(2), 'B': range(3)})
        >>> di
           A |    B |
        ==== | ==== |
        0  0 | 0  0 |
        1  1 | 1  1 |
             | 2  2 |

        >>> empty = di.copy_empty()
        >>> empty
        Empty DictOfSeries
        Columns: ['A', 'B']

        The properties are the same, eg.

        >>> empty.itype == di.itype
        True
        >>> empty.cast_policy == di.cast_policy
        True
        >>> empty.dtypes == di.dtypes
        columns
        A    True
        B    True
        dtype: bool
        """
        data = None
        if columns is True:  # is correct
            data = pd.Series(dtype="O", index=self.columns)
            for c, series in self.items():
                # OPTIM: the following code is about 2x faster than
                # data.at[c] = pd.Series(dtype=self._data.at[c].dtype)
                data.at[c] = series.reindex([])

        return self._constructor(
            data=data, itype=self.itype, cast_policy=self.cast_policy, fastpath=True
        )._finalize(self)

    # ------------------------------------------------------------------------------
    # Operators

    def _op1(self, op):
        new = self.copy_empty(columns=True)
        try:
            for k, series in self.items():
                new[k] = op(series)
        except Exception as e:
            raise type(e)(f"'{ops.OP_MAP[op]} dios' failed: " + str(e)) from e
        return new

    def _op2_inplace(self, op, other, align=True) -> None:
        new = self._op2(op, other, align)
        self._data = new._data

    def _op2(self, op, other, align=True) -> "_DiosBase":
        def raiseif(kself, kother, s):
            if kself != kother:
                raise ValueError(
                    f"{s} does not match, {s} left: {kself}, {s} right: {kother}"
                )

        def doalign(left, right):
            return left.align(right, join="inner") if align else (left, right)

        def get_operants():
            if _is_dios_like(other):
                raiseif(list(self), list(other), "keys")
                for k, series in self.items():
                    yield (k, *doalign(series, other[k]))
            elif isinstance(other, pd.Series):
                for k, series in self.items():
                    yield (k, *doalign(series, other))
            elif pdextra.is_dict_like(other):
                raiseif(sorted(self), sorted(other), "keys")
                for k, series in self.items():
                    yield (k, series, other[k])
            elif pdextra.is_nested_list_like(other):
                raiseif(len(self), len(other), "length")
                for i, (k, series) in enumerate(self.items()):
                    yield (k, series, other[i])
            elif pdextra.is_scalar(other) or pdextra.is_list_like(other):
                for k, series in self.items():
                    yield (k, series, other)
            else:
                raise NotImplementedError

        new = self.copy_empty(columns=True)
        try:
            for k, ser, oth in get_operants():
                new[k] = op(ser, oth)
        except Exception as e:
            raise type(e)(f"'dios {ops.OP_MAP[op]} other' failed: " + str(e)) from e

        return new

    # unary
    __neg__ = ftools.partialmethod(_op1, op.neg)
    __abs__ = ftools.partialmethod(_op1, op.abs)
    __invert__ = ftools.partialmethod(_op1, op.inv)
    # comparison
    __eq__ = ftools.partialmethod(_op2, op.eq, align=False)
    __ne__ = ftools.partialmethod(_op2, op.ne, align=False)
    __le__ = ftools.partialmethod(_op2, op.le, align=False)
    __ge__ = ftools.partialmethod(_op2, op.ge, align=False)
    __lt__ = ftools.partialmethod(_op2, op.lt, align=False)
    __gt__ = ftools.partialmethod(_op2, op.gt, align=False)
    # arithmetic
    __add__ = ftools.partialmethod(_op2, op.add)
    __sub__ = ftools.partialmethod(_op2, op.sub)
    __mul__ = ftools.partialmethod(_op2, op.mul)
    __mod__ = ftools.partialmethod(_op2, op.mod)
    __truediv__ = ftools.partialmethod(_op2, op.truediv)
    __floordiv__ = ftools.partialmethod(_op2, op.floordiv)
    __pow__ = ftools.partialmethod(_op2, op.pow)
    __iadd__ = ftools.partialmethod(_op2_inplace, op.add)
    __isub__ = ftools.partialmethod(_op2_inplace, op.sub)
    __imul__ = ftools.partialmethod(_op2_inplace, op.mul)
    __imod__ = ftools.partialmethod(_op2_inplace, op.mod)
    __itruediv__ = ftools.partialmethod(_op2_inplace, op.truediv)
    __ifloordiv__ = ftools.partialmethod(_op2_inplace, op.floordiv)
    __ipow__ = ftools.partialmethod(_op2_inplace, op.pow)
    # bool
    __and__ = ftools.partialmethod(_op2, op.and_)
    __or__ = ftools.partialmethod(_op2, op.or_)
    __xor__ = ftools.partialmethod(_op2, op.xor)
    __iand__ = ftools.partialmethod(_op2, op.and_, inplace=True)
    __ior__ = ftools.partialmethod(_op2, op.or_, inplace=True)
    __ixor__ = ftools.partialmethod(_op2, op.xor, inplace=True)

    # ------------------------------------------------------------------------------
    # Indexer

    @property
    def loc(self):
        """Access a group of rows and columns by label(s) or a boolean array.

        See :ref:`indexing docs <doc_indexing:Pandas-like indexing>`
        """
        return _LocIndexer(self)

    @property
    def iloc(self):
        """Purely integer-location based indexing for selection by position.

        See :ref:`indexing docs <doc_indexing:Pandas-like indexing>`
        """
        return _iLocIndexer(self)

    @property
    def aloc(self):
        """Access a group of rows and columns by label(s) or a boolean array with automatic alignment of indexers.

        See :ref:`indexing docs <doc_indexing:Special indexer .aloc>`
        """
        return _aLocIndexer(self)

    @property
    def at(self):
        """Access a single value for a row/column label pair.

        See :ref:`indexing docs <doc_indexing:Pandas-like indexing>`
        """
        return _AtIndexer(self)

    @property
    def iat(self):
        """Access a single value for a row/column pair by integer position.

        See :ref:`indexing docs <doc_indexing:Pandas-like indexing>`
        """
        return _iAtIndexer(self)


def _is_dios_like(obj) -> bool:
    # must have columns
    # columns is some kind of pd.Index
    # iter will iter through columns
    # a `in` obj check if obj is in columns
    # obj[key] will give a pd.Series
    # obj.squeeze() give pd.Series if len(obj) == 1
    return isinstance(obj, (_DiosBase, pd.DataFrame))


def _is_bool_series(obj) -> bool:
    return isinstance(obj, pd.Series) and obj.dtype == bool


def _is_bool_dios_like(obj) -> bool:
    if not _is_dios_like(obj):
        return False
    dtypes = obj.dtypes
    if (dtypes == bool).all():
        return True
    if (dtypes == "O").any():
        return obj.apply(pdextra.is_bool_indexer).all()
    return False


# keep this here to prevent cyclic import
from .indexer import _aLocIndexer, _iLocIndexer, _LocIndexer, _iAtIndexer, _AtIndexer



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Union, Tuple, Callable, Sequence, Optional
from typing_extensions import Literal
import numpy as np
import pandas as pd
from dios import DictOfSeries

from saqc.constants import *
from saqc.core import register, Flags
from saqc.lib.tools import toSequence
import saqc.lib.ts_operators as ts_ops


@register(
    mask=["field"],
    demask=[],
    squeeze=["target"],
    multivariate=True,
    handles_target=True,
)
def assignKNNScore(
    data: DictOfSeries,
    field: Sequence[str],
    flags: Flags,
    target: str,
    n: int = 10,
    func: Callable[[pd.Series], float] = np.sum,
    freq: Union[float, str] = np.inf,
    min_periods: int = 2,
    method: Literal["ball_tree", "kd_tree", "brute", "auto"] = "ball_tree",
    metric: str = "minkowski",
    p: int = 2,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    TODO: docstring need a rework
    Score datapoints by an aggregation of the dictances to their k nearest neighbors.

    The function is a wrapper around the NearestNeighbors method from pythons sklearn library (See reference [1]).

    The steps taken to calculate the scores are as follows:

    1. All the timeseries, given through ``field``, are combined to one feature space by an *inner* join on their
       date time indexes. thus, only samples, that share timestamps across all ``field`` will be included in the
       feature space.
    2. Any datapoint/sample, where one ore more of the features is invalid (=np.nan) will get excluded.
    3. For every data point, the distance to its `n` nearest neighbors is calculated by applying the
       metric `metric` at grade `p` onto the feature space. The defaults lead to the euclidian to be applied.
       If `radius` is not None, it sets the upper bound of distance for a neighbor to be considered one of the
       `n` nearest neighbors. Furthermore, the `freq` argument determines wich samples can be
       included into a datapoints nearest neighbors list, by segmenting the data into chunks of specified temporal
       extension and feeding that chunks to the kNN algorithm seperatly.
    4. For every datapoint, the calculated nearest neighbors distances get aggregated to a score, by the function
       passed to `func`. The default, ``sum`` obviously just sums up the distances.
    5. The resulting timeseries of scores gets assigned to the field target.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : list of str
        input variable names.
    flags : saqc.flags
        A flags object, holding flags and additional informations related to `data`.
    target : str, default "kNNscores"
        A new Column name, where the result is stored.
    n : int, default 10
        The number of nearest neighbors to which the distance is comprised in every datapoints scoring calculation.
    func : Callable[numpy.array, float], default np.sum
        A function that assigns a score to every one dimensional array, containing the distances
        to every datapoints `n` nearest neighbors.
    freq : {np.inf, float, str}, default np.inf
        Determines the segmentation of the data into partitions, the kNN algorithm is
        applied onto individually.

        * ``np.inf``: Apply Scoring on whole data set at once
        * ``x`` > 0 : Apply scoring on successive data chunks of periods length ``x``
        * Offset String : Apply scoring on successive partitions of temporal extension matching the passed offset
          string

    min_periods : int, default 2
        The minimum number of periods that have to be present in a partition for the kNN scoring
        to be applied. If the number of periods present is below `min_periods`, the score for the
        datapoints in that partition will be np.nan.
    method : {'ball_tree', 'kd_tree', 'brute', 'auto'}, default 'ball_tree'
        The search algorithm to find each datapoints k nearest neighbors.
        The keyword just gets passed on to the underlying sklearn method.
        See reference [1] for more information on the algorithm.
    metric : str, default 'minkowski'
        The metric the distances to any datapoints neighbors is computed with. The default of `metric`
        together with the default of `p` result in the euclidian to be applied.
        The keyword just gets passed on to the underlying sklearn method.
        See reference [1] for more information on the algorithm.
    p : int, default 2
        The grade of the metrice specified by parameter `metric`.
        The keyword just gets passed on to the underlying sklearn method.
        See reference [1] for more information on the algorithm.

    References
    ----------
    [1] https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.NearestNeighbors.html
    """
    if target in data.columns:
        raise ValueError(f"'target' must not exist.")
    fields = toSequence(field)
    val_frame = data[fields].copy()
    score_index = val_frame.index_of("shared")
    score_ser = pd.Series(np.nan, index=score_index, name=target)

    val_frame = val_frame.loc[val_frame.index_of("shared")].to_df()
    val_frame.dropna(inplace=True)

    if val_frame.empty:
        return data, flags

    # partitioning
    if not freq:
        freq = val_frame.shape[0]

    if isinstance(freq, str):
        grouper = pd.Grouper(freq=freq)
    else:
        grouper = pd.Series(
            data=np.arange(0, val_frame.shape[0]), index=val_frame.index
        )
        grouper = grouper.transform(lambda x: int(np.floor(x / freq)))

    partitions = val_frame.groupby(grouper)

    for _, partition in partitions:
        if partition.empty or (partition.shape[0] < min_periods):
            continue

        sample_size = partition.shape[0]
        nn_neighbors = min(n, max(sample_size, 2) - 1)
        dist, *_ = ts_ops.kNN(
            partition.values, nn_neighbors, algorithm=method, metric=metric, p=p
        )
        try:
            resids = getattr(dist, func.__name__)(axis=1)
        except AttributeError:
            resids = np.apply_along_axis(func, 1, dist)

        score_ser[partition.index] = resids

    flags[target] = pd.Series(UNFLAGGED, index=score_ser.index, dtype=float)
    data[target] = score_ser

    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations
from typing import Union

from dios import DictOfSeries
from typing_extensions import Literal

from saqc.constants import BAD
import saqc


class Curvefit:
    def fitPolynomial(
        self,
        field: str,
        window: Union[int, str],
        order: int,
        min_periods: int = 0,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("fitPolynomial", locals())



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Callable, Optional, Union

import pandas as pd

import saqc


class Transformation:
    def transform(
        self,
        field: str,
        func: Callable[[pd.Series], pd.Series],
        freq: Optional[Union[float, str]] = None,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("transform", locals())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from math import floor
from typing import Tuple, Union
from typing_extensions import Literal
import numpy as np
import pandas as pd
from dios import DictOfSeries

from saqc.constants import *
from saqc.core import register, Flags
from saqc.lib.tools import getFreqDelta, filterKwargs
from saqc.lib.ts_operators import (
    polyRollerIrregular,
    polyRollerNumba,
    polyRoller,
    polyRollerNoMissingNumba,
    polyRollerNoMissing,
)


@register(mask=["field"], demask=[], squeeze=[])
def fitPolynomial(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: Union[int, str],
    order: int,
    min_periods: int = 0,
    **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Fits a polynomial model to the data.

    The fit is calculated by fitting a polynomial of degree `order` to a data slice
    of size `window`, that has x at its center.

    Note that the result is stored in `field` and overwrite it unless a
    `target` is given.

    In case your data is sampled at an equidistant frequency grid:

    (1) If you know your data to have no significant number of missing values,
    or if you do not want to calculate residues for windows containing missing values
    any way, performance can be increased by setting min_periods=window.

    Note, that the initial and final window/2 values do not get fitted.

    Each residual gets assigned the worst flag present in the interval of
    the original data.

    Parameters
    ----------
    data : dios.DictOfSeries
        The data.

    field : str
        The column, holding the data-to-be-modelled.

    flags : saqc.Flags
        Container to store quality flags to data.

    window : {str, int}
        The size of the window you want to use for fitting. If an integer is passed,
        the size refers to the number of periods for every fitting window. If an
        offset string is passed, the size refers to the total temporal extension. The
        window will be centered around the vaule-to-be-fitted. For regularly sampled
        timeseries the period number will be casted down to an odd number if even.

    order : int
        The degree of the polynomial used for fitting

    min_periods : int or None, default 0
        The minimum number of periods, that has to be available in every values
        fitting surrounding for the polynomial fit to be performed. If there are not
        enough values, np.nan gets assigned. Default (0) results in fitting
        regardless of the number of values present (results in overfitting for too
        sparse intervals). To automatically set the minimum number of periods to the
        number of values in an offset defined window size, pass np.nan.

    Returns
    -------
    data : dios.DictOfSeries
    flags : saqc.Flags
    """
    reserved = ["residues", "set_flags"]
    filterKwargs(kwargs, reserved)
    return _fitPolynomial(
        data=data,
        field=field,
        flags=flags,
        window=window,
        order=order,
        min_periods=min_periods,
        **kwargs,
        # ctrl args
        return_residues=False,
        set_flags=True,
    )


def _fitPolynomial(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: Union[int, str],
    order: int,
    set_flags: bool = True,
    min_periods: int = 0,
    return_residues: bool = False,
    **kwargs
) -> Tuple[DictOfSeries, Flags]:

    # TODO: some (rather large) parts are functional similar to saqc.funcs.rolling.roll
    if data[field].empty:
        return data, flags

    to_fit = data[field].copy()
    regular = getFreqDelta(to_fit.index)
    if not regular:
        if isinstance(window, int):
            raise NotImplementedError(
                "Integer based window size is not supported for not-harmonized"
                "sample series."
            )
        # get interval centers
        centers = (
            to_fit.rolling(
                pd.Timedelta(window) / 2, closed="both", min_periods=min_periods
            ).count()
        ).floor()
        centers = centers.drop(centers[centers.isna()].index)
        centers = centers.astype(int)
        residues = to_fit.rolling(
            pd.Timedelta(window), closed="both", min_periods=min_periods
        ).apply(polyRollerIrregular, args=(centers, order))

        def center_func(x, y=centers):
            pos = x.index[int(len(x) - y[x.index[-1]])]
            return y.index.get_loc(pos)

        centers_iloc = (
            centers.rolling(window, closed="both")
            .apply(center_func, raw=False)
            .astype(int)
        )
        temp = residues.copy()
        for k in centers_iloc.iteritems():
            residues.iloc[k[1]] = temp[k[0]]
        residues[residues.index[0] : residues.index[centers_iloc[0]]] = np.nan
        residues[residues.index[centers_iloc[-1]] : residues.index[-1]] = np.nan
    else:
        if isinstance(window, str):
            window = pd.Timedelta(window) // regular
        if window % 2 == 0:
            window = int(window - 1)
        if min_periods is None:
            min_periods = window
        if to_fit.shape[0] < 200000:
            numba = False
        else:
            numba = True

        val_range = np.arange(0, window)
        center_index = window // 2
        if min_periods < window:
            if min_periods > 0:
                to_fit = to_fit.rolling(
                    window, min_periods=min_periods, center=True
                ).apply(lambda x, y: x[y], raw=True, args=(center_index,))

            # we need a missing value marker that is not nan,
            # because nan values dont get passed by pandas rolling method
            miss_marker = to_fit.min()
            miss_marker = np.floor(miss_marker - 1)
            na_mask = to_fit.isna()
            to_fit[na_mask] = miss_marker
            if numba:
                residues = to_fit.rolling(window).apply(
                    polyRollerNumba,
                    args=(miss_marker, val_range, center_index, order),
                    raw=True,
                    engine="numba",
                    engine_kwargs={"no_python": True},
                )
                # due to a tiny bug - rolling with center=True doesnt work
                # when using numba engine.
                residues = residues.shift(-int(center_index))
            else:
                residues = to_fit.rolling(window, center=True).apply(
                    polyRoller,
                    args=(miss_marker, val_range, center_index, order),
                    raw=True,
                )
            residues[na_mask] = np.nan
        else:
            # we only fit fully populated intervals:
            if numba:
                residues = to_fit.rolling(window).apply(
                    polyRollerNoMissingNumba,
                    args=(val_range, center_index, order),
                    engine="numba",
                    engine_kwargs={"no_python": True},
                    raw=True,
                )
                # due to a tiny bug - rolling with center=True doesnt work
                # when using numba engine.
                residues = residues.shift(-int(center_index))
            else:
                residues = to_fit.rolling(window, center=True).apply(
                    polyRollerNoMissing,
                    args=(val_range, center_index, order),
                    raw=True,
                )

    if return_residues:
        residues = to_fit - residues

    data[field] = residues
    if set_flags:
        # TODO: we does not get any flags here, because of masking=field
        worst = flags[field].rolling(window, center=True, min_periods=min_periods).max()
        flags[field] = worst

    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Optional, Tuple

from typing_extensions import Literal
import numpy as np
from dios import DictOfSeries

import matplotlib as mpl
import matplotlib.pyplot as plt
import pickle

from saqc.constants import *
from saqc.core.register import processing
from saqc.core import register, Flags
from saqc.lib.tools import periodicMask, filterKwargs
from saqc.lib.plotting import makeFig

_MPL_DEFAULT_BACKEND = mpl.get_backend()


@register(mask=[], demask=[], squeeze=[], handles_target=True)
def copyField(
    data: DictOfSeries, field: str, flags: Flags, target: str, **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Copy data and flags to a new name (preserve flags history).

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the data column, you want to fork (copy).
    flags : saqc.Flags
        Container to store quality flags to data.
    target: str
        Target name.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        data shape may have changed relatively to the flags input.
    flags : saqc.Flags
        The quality flags of data
        Flags shape may have changed relatively to the flags input.
    """
    if target in flags.columns.union(data.columns):
        raise ValueError(f"{field}: field already exist")

    data[target] = data[field].copy()
    flags.history[target] = flags.history[field].copy()

    return data, flags


@processing()
def dropField(
    data: DictOfSeries, field: str, flags: Flags, **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Drops field from the data and flags.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the data column, you want to drop.
    flags : saqc.Flags
        Container to store quality flags to data.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        data shape may have changed relatively to the flags input.
    flags : saqc.Flags
        The quality flags of data
        Flags shape may have changed relatively to the flags input.
    """
    del data[field]
    del flags[field]
    return data, flags


@processing()
def renameField(
    data: DictOfSeries, field: str, flags: Flags, new_name: str, **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Rename field in data and flags.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the data column, you want to rename.
    flags : saqc.Flags
        Container to store flags of the data.
    new_name : str
        String, field is to be replaced with.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
    """
    data[new_name] = data[field]
    flags.history[new_name] = flags.history[field]
    del data[field]
    del flags[field]
    return data, flags


@register(mask=[], demask=[], squeeze=["field"])
def maskTime(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    mode: Literal["periodic", "mask_field"],
    mask_field: Optional[str] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    closed: bool = True,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Realizes masking within saqc.

    Due to some inner saqc mechanics, it is not straight forwardly possible to exclude
    values or datachunks from flagging routines. This function replaces flags with UNFLAGGED
    value, wherever values are to get masked. Furthermore, the masked values get replaced by
    np.nan, so that they dont effect calculations.

    Here comes a recipe on how to apply a flagging function only on a masked chunk of the variable field:

    1. dublicate "field" in the input data (`copyField`)
    2. mask the dublicated data (this, `maskTime`)
    3. apply the tests you only want to be applied onto the masked data chunks (a saqc function)
    4. project the flags, calculated on the dublicated and masked data onto the original field data
        (`concateFlags` or `flagGeneric`)
    5. drop the dublicated data (`dropField`)

    To see an implemented example, checkout flagSeasonalRange in the saqc.functions module

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-masked.
    flags : saqc.Flags
        Container to store flags of the data.
    mode : {"periodic", "mask_var"}
        The masking mode.
        - "periodic": parameters "period_start", "end" are evaluated to generate a periodical mask
        - "mask_var": data[mask_var] is expected to be a boolean valued timeseries and is used as mask.
    mask_field : {None, str}, default None
        Only effective if mode == "mask_var"
        Fieldname of the column, holding the data that is to be used as mask. (must be boolean series)
        Neither the series` length nor its labels have to match data[field]`s index and length. An inner join of the
        indices will be calculated and values get masked where the values of the inner join are ``True``.
    start : {None, str}, default None
        Only effective if mode == "seasonal"
        String denoting starting point of every period. Formally, it has to be a truncated instance of "mm-ddTHH:MM:SS".
        Has to be of same length as `end` parameter.
        See examples section below for some examples.
    end : {None, str}, default None
        Only effective if mode == "periodic"
        String denoting starting point of every period. Formally, it has to be a truncated instance of "mm-ddTHH:MM:SS".
        Has to be of same length as `end` parameter.
        See examples section below for some examples.
    closed : boolean
        Wheather or not to include the mask defining bounds to the mask.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed relatively to the flags input.


    Examples
    --------
    The `period_start` and `end` parameters provide a conveniant way to generate seasonal / date-periodic masks.
    They have to be strings of the forms: "mm-ddTHH:MM:SS", "ddTHH:MM:SS" , "HH:MM:SS", "MM:SS" or "SS"
    (mm=month, dd=day, HH=hour, MM=minute, SS=second)
    Single digit specifications have to be given with leading zeros.
    `period_start` and `seas   on_end` strings have to be of same length (refer to the same periodicity)
    The highest date unit gives the period.
    For example:

    >>> period_start = "01T15:00:00"
    >>> end = "13T17:30:00"

    Will result in all values sampled between 15:00 at the first and  17:30 at the 13th of every month get masked

    >>> period_start = "01:00"
    >>> end = "04:00"

    All the values between the first and 4th minute of every hour get masked.

    >>> period_start = "01-01T00:00:00"
    >>> end = "01-03T00:00:00"

    Mask january and february of evcomprosed in theery year. masking is inclusive always, so in this case the mask will
    include 00:00:00 at the first of march. To exclude this one, pass:

    >>> period_start = "01-01T00:00:00"
    >>> end = "02-28T23:59:59"

    To mask intervals that lap over a seasons frame, like nights, or winter, exchange sequence of season start and
    season end. For example, to mask night hours between 22:00:00 in the evening and 06:00:00 in the morning, pass:

    >>> period_start = "22:00:00"
    >>> end = "06:00:00"

    When inclusive_selection="season", all above examples work the same way, only that you now
    determine wich values NOT TO mask (=wich values are to constitute the "seasons").
    """
    datcol_idx = data[field].index

    if mode == "periodic":
        mask = periodicMask(datcol_idx, start, end, closed)
    elif mode == "mask_field":
        idx = data[mask_field].index.intersection(datcol_idx)
        mask = data.loc[idx, mask_field]
    else:
        raise ValueError("Keyword passed as masking mode is unknown ({})!".format(mode))

    data.aloc[mask, field] = np.nan
    flags[mask, field] = UNFLAGGED
    return data, flags


@register(mask=[], demask=[], squeeze=[])
def plot(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    path: Optional[str] = None,
    max_gap: Optional[str] = None,
    stats: bool = False,
    history: Optional[Literal["valid", "complete", "clear"]] = "valid",
    xscope: Optional[slice] = None,
    phaseplot: Optional[str] = None,
    stats_dict: Optional[dict] = None,
    store_kwargs: Optional[dict] = None,
    dfilter: float = FILTER_ALL,
    **kwargs,
):
    """
    Plot data and flags or store plot to file.

    There are two modes, 'interactive' and 'store', which are determind through the
    ``save_path`` keyword. In interactive mode (default) the plot is shown at runtime
    and the program execution stops until the plot window is closed manually. In
    store mode the generated plot is stored to disk and no manually interaction is
    needed.

    Parameters
    ----------
    data : {pd.DataFrame, dios.DictOfSeries}
        data

    field : str
        Name of the variable-to-plot

    flags : {pd.DataFrame, dios.DictOfSeries, saqc.flagger}
        Flags or flagger object

    path : str, default None
        If ``None`` is passed, interactive mode is entered; plots are shown immediatly
        and a user need to close them manually before execution continues.
        If a filepath is passed instead, store-mode is entered and
        the plot is stored unter the passed location.

    max_gap : str, default None
        If None, all the points in the data will be connected, resulting in long linear
        lines, where continous chunks of data is missing. Nans in the data get dropped
        before plotting. If an offset string is passed, only points that have a distance
        below `max_gap` get connected via the plotting line.

    stats : bool, default False
        Whether to include statistics table in plot.

    history : {"valid", "complete", None}, default "valid"
        Discriminate the plotted flags with respect to the tests they originate from.

        * "valid" - Only plot those flags, that do not get altered or "unflagged" by subsequent tests. Only list tests
          in the legend, that actually contributed flags to the overall resault.
        * "complete" - plot all the flags set and list all the tests ran on a variable. Suitable for debugging/tracking.
        * "clear" - clear plot from all the flagged values
        * None - just plot the resulting flags for one variable, without any historical meta information.

    xscope : slice or Offset, default None
        Parameter, that determines a chunk of the data to be plotted
        processed. `xscope` can be anything, that is a valid argument to the ``pandas.Series.__getitem__`` method.

    phaseplot : str or None, default None
        If a string is passed, plot ``field`` in the phase space it forms together with the Variable ``phaseplot``.

    store_kwargs : dict, default {}
        Keywords to be passed on to the ``matplotlib.pyplot.savefig`` method, handling
        the figure storing. To store an pickle object of the figure, use the option
        ``{'pickle': True}``, but note that all other store_kwargs are ignored then.
        Reopen with: ``pickle.load(open(savepath,'w')).show()``

    stats_dict: dict, default None
        (Only relevant if ``stats = True``)
        Dictionary of additional statisticts to write to the statistics table
        accompanying the data plot. An entry to the stats_dict has to be of the form:

        * ``{"stat_name": lambda x, y, z: func(x, y, z)}``

        The lambda args ``x``,``y``,``z`` will be fed by:

        * ``x``: the data (``data[field]``).
        * ``y``: the flags (``flags[field]``).
        * ``z``: The passed flags level (``kwargs[flag]``)

        See examples section for examples

    Examples
    --------
    Summary statistic function examples:

    >>> func = lambda x, y, z: len(x)

    Total number of nan-values:

    >>> func = lambda x, y, z: x.isna().sum()

    Percentage of values, flagged greater than passed flag (always round float results
    to avoid table cell overflow):

    >>> func = lambda x, y, z: round((x.isna().sum()) / len(x), 2)
    """

    interactive = path is None
    level = kwargs.get("flag", BAD)

    if dfilter < np.inf:
        data = data.copy()
        data.loc[flags[field] >= dfilter, field] = np.nan

    if store_kwargs is None:
        store_kwargs = {}

    if interactive:
        mpl.use(_MPL_DEFAULT_BACKEND)

    else:
        mpl.use("Agg")

    fig = makeFig(
        data=data,
        field=field,
        flags=flags,
        level=level,
        max_gap=max_gap,
        stats=stats,
        history=history,
        xscope=xscope,
        phaseplot=phaseplot,
        stats_dict=stats_dict,
    )

    if interactive:
        plt.show()

    else:
        if store_kwargs.pop("pickle", False):
            with open(path, "wb") as f:
                pickle.dump(fig, f)
        else:
            fig.savefig(path, **store_kwargs)

    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from saqc.core.modules.breaks import Breaks
from saqc.core.modules.noise import Noise
from saqc.core.modules.changepoints import ChangePoints
from saqc.core.modules.constants import Constants
from saqc.core.modules.curvefit import Curvefit
from saqc.core.modules.drift import Drift
from saqc.core.modules.flagtools import FlagTools
from saqc.core.modules.generic import Generic
from saqc.core.modules.interpolation import Interpolation
from saqc.core.modules.outliers import Outliers
from saqc.core.modules.pattern import Pattern
from saqc.core.modules.resampling import Resampling
from saqc.core.modules.residues import Residues
from saqc.core.modules.rolling import Rolling
from saqc.core.modules.scores import Scores
from saqc.core.modules.tools import Tools
from saqc.core.modules.transformation import Transformation
from saqc.core.register import FUNC_MAP


class FunctionsMixin(
    Breaks,
    Noise,
    ChangePoints,
    Constants,
    Curvefit,
    Drift,
    FlagTools,
    Generic,
    Interpolation,
    Outliers,
    Pattern,
    Resampling,
    Residues,
    Rolling,
    Scores,
    Tools,
    Transformation,
):
    def _defer(self, fname, flocals):
        flocals.pop("self", None)
        fkwargs = flocals.pop("kwargs", {})
        return self._wrap(FUNC_MAP[fname])(**flocals, **fkwargs)



from __future__ import annotations
from typing import Any, Mapping, Hashable

from .base import _DiosBase, _is_dios_like
from .lib import Opts, OptsFields, dios_options
from .lib import _find_least_common_itype
from . import pandas_bridge as pdextra

import functools as ftools
import pandas as pd
import numpy as np


class DictOfSeries(_DiosBase):
    """A data frame where every column has its own index.

    DictOfSeries is a collection of pd.Series's which aim to be as close as possible similar to
    pd.DataFrame. The advantage over pd.DataFrame is, that every `column` has its own row-index,
    unlike the former, which provide a single row-index for all columns. This solves problems with
    unaligned data and data which varies widely in length.

    Indexing with ``di[]``, ``di.loc[]`` and ``di.iloc[]``  should work analogous to these methods
    from pd.DataFrame. The indexer can be a single label, a slice, a list-like, a boolean list-like,
    or a boolean DictOfSeries/pd.DataFrame and can be used to selectively get or set data.

    Parameters
    ----------
    data : array-like, Iterable, dict, or scalar value
        Contains data stored in Series.

    columns : array-like
        Column labels to use for resulting frame. Will default to
        RangeIndex(0, 1, 2, ..., n) if no column labels are provided.

    index : Index or array-like
        Index to use to reindex every given series during init. Ignored if omitted.

    itype : Itype, pd.Index, Itype-string-repr or type
        Every series that is inserted, must have an index of this type or any
        of this types subtypes.
        If None, the itype is inferred as soon as the first non-empty series is inserted.

    cast_policy : {'save', 'force', 'never'}, default 'save'
        Policy used for (down-)casting the index of a series if its type does not match
        the ``itype``.
    """

    @property
    def _constructor(self) -> type[DictOfSeries]:
        """Return the class. Useful for construction in the elder class.
        A import of DictOfSeries would end up cyclic."""
        return DictOfSeries

    def _construct_like_self(self, **kwargs):
        kwargs.setdefault("itype", self.itype)
        kwargs.setdefault("cast_policy", self.cast_policy)
        return self._constructor(**kwargs)._finalize(self)

    @property
    def indexes(self):
        """Return pandas.Series with the indexes of all columns."""
        return self.for_each("index")

    @property
    def values(self):
        """Return a numpy.array of numpy.arrays with the values of all columns.

        The outer has the length of columns, the inner holds the values of the column.
        """
        s = self.for_each("values")
        return s.values

    @property
    def dtypes(self):
        """Return pandas.Series with the dtypes of all columns."""
        return self.for_each("dtype")

    @property
    def lengths(self):
        """Return pandas.Series with the lenght of all columns."""
        return self._data.apply(len)

    @property
    def size(self):
        return self.lengths.sum()

    @property
    def shape(self):
        return tuple(self.lengths), len(self.columns)

    # ------------------------------------------------------------------------------
    # Dict-like methods

    def clear(self):
        d = self._data
        self._data = pd.Series(dtype=d.dtype, index=type(d.index)([]))

    def get(self, key, default=None):
        return self._data.get(key, default)

    # implemented in _BaseClass
    # def items(self):
    #     return self._data.items()

    def keys(self):
        return self.columns

    def pop(self, *args):
        # We support a default value, like dict, in contrary to pd.
        # Therefore we need to handle args manually, because dict-style pop()
        # differ between a single arg and a tuple-arg, with arg and default,
        # where the second arg can be anything, including None. If the key is
        # not present, and a single arg is given, a KeyError is raised, but
        # with a given default value, it is returned instead.
        if len(args) == 0:
            raise TypeError("pop expected at least 1 arguments, got 0")
        if len(args) > 2:
            raise TypeError(f"pop expected at most 2 arguments, got {len(args)}")
        key, *rest = args
        if key in self.columns:
            return self._data.pop(key)
        elif rest:
            return rest.pop()
        raise KeyError(key)

    def popitem(self):
        last = self.columns[-1]
        return last, self._data.pop(last)

    def setdefault(self, key, default=None):
        if key not in self.columns:
            self._insert(key, default)
        return self._data[key]

    def update(self, other):
        if not _is_dios_like(other):
            other = to_dios(other)
        self.aloc[other, ...] = other

    # ------------------------------------------------------------------------------
    # High-Level Iteration

    def iteritems(self):
        yield from self.items()

    def iterrows(self, fill_value=np.nan, squeeze=True):
        """
        Iterate over DictOfSeries rows as (index, pandas.Series/DictOfSeries) pairs.
        **MAY BE VERY PERFORMANCE AND/OR MEMORY EXPENSIVE**

        Parameters
        ----------
        fill_value: scalar, default numpy.nan
            Fill value for row entry, if the column does not have an entry
            at the current index location. This ensures that the returned
            Row always contain all columns. If ``None`` is given no value
            is filled.

            If ``fill_value=None`` and ``squeeze=True`` the resulting Row
            (a pandas.Series) may differ in length between iterator calls.
            That's because an entry, that is not present in a column, will
            also not be present in the resulting Row.

        squeeze: bool, default False
            * ``True`` : A pandas.Series is returned for each row.
            * ``False`` : A single-rowed DictOfSeries is returned for each row.

        Yields
        ------
        index : label
            The index of the row.
        data : Series or DictOfSeries
            The data of the row as a Series if squeeze is True, as
            a DictOfSeries otherwise.

        See Also
        --------
        DictOfSeries.iteritems : Iterate over (column name, Series) pairs.
        """

        # todo: 2nd posibility for fill_value=Any, squeeze=False
        #   do it like in case fill_value=None ->
        #       1. row = aloc the row
        #       2. e = row.isempty()
        #       3. row.loc[idx,e] = fill_value
        #   This approach could be much better, because the dtype of
        #   the columns is preserved.

        # PROBABLY PERFORMANCE EXPENSIVE
        if fill_value is None:
            allidx = self.index_of("all")
            if squeeze:
                for i in allidx:
                    yield i, self.aloc[i:i].dropempty().squeeze(axis=0)
            else:
                for i in allidx:
                    yield self.aloc[i:i]

        # PROBABLY MEMORY EXPENSIVE
        else:
            if fill_value is np.nan:
                df = self.to_df()
            else:
                nans = self.isna().to_df().fillna(False)
                df = self.to_df().fillna(fill_value)
                df[nans] = np.nan
            if squeeze:
                yield from df.iterrows()
            else:
                for idx, row in df.iterrows():
                    yield idx, self._constructor(
                        data=row.to_dict(), index=[idx]
                    )._finalize(self)

    # ------------------------------------------------------------------------------
    # Broadcasting and Reducing

    def for_each(self, attr_or_callable, **kwds):
        """
        Apply a callable or a pandas.Series method or property on each column.

        Parameters
        ----------
        attr_or_callable: Any
            A pandas.Series attribute or any callable, to apply on each column.
            A series attribute can be any property, field or method and also
            could be specified as string. If a callable is given it must take
            pandas.Series as the only positional argument and return a scalar.

        **kwds: any
            kwargs to passed to callable

        Returns
        -------
        pandas.Series
            A series with the results, indexed by the column labels.

        Notes
        -----
        The called function or the attribute works on the actual underlying series.
        If the provided function works inplace it can and will modify the actual data.
        If this is not desired one can should make an explicit copy beforehand. If the
        function returns new objects or copies, explicit copying is not needed.

        See Also
        --------
        DictOfSeries.apply : Apply functions to columns and convert
                             result to DictOfSeries.

        Examples
        --------
        >>> d = DictOfSeries([range(3), range(4)], columns=['a', 'b'])
        >>> d
           a |    b |
        ==== | ==== |
        0  0 | 0  0 |
        1  1 | 1  1 |
        2  2 | 2  2 |
             | 3  3 |

        Use with a callable..

        >>> d.for_each(max)
        columns
        a    2
        b    3
        dtype: object

        ..or with a string, denoting a pd.Series attribute and
        therefor is the same as giving the latter.

        >>> d.for_each('max')
        columns
        a    2
        b    3
        dtype: object

        >>> d.for_each(pd.Series.max)
        columns
        a    2
        b    3
        dtype: object

        Both also works with properties:

        >>> d.for_each('dtype')
        columns
        a    int64
        b    int64
        dtype: object
        """
        attrOcall = attr_or_callable
        if isinstance(attrOcall, str):
            attrOcall = getattr(pd.Series, attrOcall)
        call = callable(attrOcall)
        if not call:
            attrOcall = attr_or_callable
        data = pd.Series(dtype="O", index=self.columns)
        for c, series in self.items():
            if call:
                data.at[c] = attrOcall(series, **kwds)
            else:
                data.at[c] = getattr(series, attrOcall)
        return data

    def apply(self, func, axis=0, raw=False, args=(), **kwds):
        """
        Apply a function along an axis of the DictOfSeries.

        Parameters
        ----------
        func : callable
            Function to apply on each column.
        axis : {0 or 'index', 1 or 'columns'}, default 0
            Axis along which the function is applied:

            * 0 or 'index': apply function to each column.
            * 1 or 'columns': NOT IMPLEMENTED

        raw : bool, default False
            Determines if row or column is passed as a Series or ndarray object:

            * ``False`` : passes each row or column as a Series to the
              function.
            * ``True`` : the passed function will receive ndarray objects
              instead.
              If you are just applying a NumPy reduction function this will
              achieve much better performance.

        args : tuple
            Positional arguments to pass to `func` in addition to the
            array/series.
        **kwds
            Additional keyword arguments to pass as keywords arguments to
            `func`.

        Returns
        -------
        Series or DataFrame
            Result of applying ``func`` along the given axis of the
            DataFrame.

        Notes
        -----
        The called function or the attribute works on the actual underlying series.
        If the provided function works inplace it can and will modify the actual data.
        If this is not desired one should make an explicit copy beforehand. If the
        function returns new objects or copies, and does not mess with the data, explicit
        copying is not needed.


        Raises
        ------
        NotImplementedError
            * if axis is 'columns' or 1

        See Also
        --------
        DictOfSeries.for_each: apply pd.Series methods or properties to each column

        Examples
        --------

        We use the example DictOfSeries from :ref:`indexing <doc_indexing:Example dios>`.

        >>> di = di[:5]
            a |    b |     c |     d |
        ===== | ==== | ===== | ===== |
        0   0 | 2  5 | 4   7 | 6   0 |
        1   7 | 3  6 | 5  17 | 7   1 |
        2  14 | 4  7 | 6  27 | 8   2 |
        3  21 | 5  8 | 7  37 | 9   3 |
        4  28 | 6  9 | 8  47 | 10  4 |

        >>> di.apply(max)
        columns
        a    28
        b     9
        c    47
        d     4
        dtype: int64

        >>> di.apply(pd.Series.count)
        columns
        a    5
        b    5
        c    5
        d    5
        dtype: int64

        One can pass keyword arguments directly..

        >>> di.apply(pd.Series.value_counts, normalize=True)
              a |      b |       c |      d |
        ======= | ====== | ======= | ====== |
        7   0.2 | 7  0.2 | 7   0.2 | 4  0.2 |
        14  0.2 | 6  0.2 | 37  0.2 | 3  0.2 |
        21  0.2 | 5  0.2 | 47  0.2 | 2  0.2 |
        28  0.2 | 9  0.2 | 27  0.2 | 1  0.2 |
        0   0.2 | 8  0.2 | 17  0.2 | 0  0.2 |

        Or define a own funtion..

        >>> di.apply(lambda s : 'high' if max(s) > 10 else 'low')
        columns
        a    high
        b     low
        c    high
        d     low
        dtype: object

        And also more advanced functions that return a list-like can be given. Note that
        the returned lists not necessarily must have the same length.

        >>> func = lambda s : ('high', max(s), min(s)) if min(s) > (max(s)//2) else ('low',max(s))
        >>> di.apply(func)
             a |       b |      c |      d |
        ====== | ======= | ====== | ====== |
        0  low | 0  high | 0  low | 0  low |
        1   28 | 1     9 | 1   47 | 1    4 |
               | 2     5 |        |        |
        """
        if axis in [1, "columns"]:
            raise NotImplementedError

        if axis not in [0, "index"]:
            raise ValueError(axis)

        # we cannot use self._data.apply(func=func, args=args, **kwds)
        # because this may return a pandas.DataFrame. Also we cannot
        # use pandas.Series.apply(), because this works on its values.
        need_dios = need_convert = False
        result = pd.Series(dtype="O", index=self.columns)
        for c, series in self.items():
            series = series.values if raw else series
            s = func(series, *args, **kwds)
            result.at[c] = s
            if pdextra.is_scalar(s):
                need_convert = True
            else:
                need_dios = True
                if not isinstance(s, pd.Series):
                    need_convert = True
        if need_dios:
            if need_convert:
                for c, val in result.items():
                    result.at[c] = pd.Series(val)
            itype = _find_least_common_itype(result)
            result = self._constructor(data=result, itype=itype, fastpath=True)
            result._finalize(self)

        return result

    def reduce_columns(self, func, initial=None, skipna=False):
        """
        Reduce all columns to a single pandas.Series by a given function.

        Apply a function of two pandas.Series as arguments, cumulatively to all
        columns, from left to right, so as to reduce the columns to a single
        pandas.Series. If initial is present, it is placed before the columns
        in the calculation, and serves as a default when the columns are empty.

        Parameters
        ----------
        func : function
            The function must take two identically indexed pandas.Series and should
            return a single pandas.Series with the same index.

        initial : column-label or pd.Series, default None
            The series to start with. If None a dummy series is created, with the
            indices of all columns and the first seen values.

        skipna : bool, default False
               If True, skip NaN values.

        Returns
        -------
        pandas.Series
            A series with the reducing result and the index of the start series,
            defined by ``initializer``.
        """
        if initial is None:
            value = pd.Series(index=self.index_of("all"))
            for d in self._data:
                value = value.combine_first(d)
        elif isinstance(initial, pd.Series):
            value = initial.copy()
        elif initial in self.columns:
            value = self._data.at[initial].copy()
        else:
            raise ValueError("initial must be pd.Series, a column label or None")

        if skipna:
            val = value.dropna()
            data = self.dropna()._data
        else:
            val = value
            data = self._data

        for d in data:
            idx = val.index & d.index
            if len(idx) > 0:
                l, r = val.loc[idx], d.loc[idx]
                val.loc[idx] = func(l, r)

        if skipna:
            value.loc[val.index] = val
        return value

    # ------------------------------------------------------------------------------
    # Merging and Joining

    def combine_first(self, other, keepna=False):
        """
        Update null elements with value in the same location in other.

        Combine two DictOfSeries objects by filling null values in one DictOfSeries with
        non-null values from other DictOfSeries. The row and column indexes of the resulting
        DictOfSeries will be the union of the two.

        Parameters
        ----------
        keepna : bool, default False
            By default Nan's are updated by other and new value-index pairs from other are
            inserted. If set to True, NaN's are not updated and only new value-index pair are inserted.

        other : DictOfSeries
            Provided DictOfSeries to use to fill null values.

        Returns
        -------
        DictOfSeries
        """
        if keepna:
            nans = self.isna()

        new: DictOfSeries = self.copy()
        for c in other.columns:
            if c in self.columns:
                col = self._data.at[c].combine_first(other[c])
            else:
                col = other[c]
            new._data.at[c] = col

        if keepna:
            new.aloc[nans] = np.nan

        return new

    # ------------------------------------------------------------------------------
    # Misc methods

    def index_of(self, method="all"):
        """Return an single index with indices from all columns.

        Parameters
        ----------
        method : string, default 'all'
            * 'all' : get all indices from all columns
            * 'union' : alias for 'all'
            * 'shared' : get indices that are present in every columns
            * 'intersection' : alias for 'shared'
            * 'uniques' : get indices that are only present in a single column
            * 'non-uniques' : get indices that are present in more than one column

        Returns
        -------
        pd.Index
            A single duplicate-free index, somehow representing indices of all columns.

        Examples
        --------
        We use the example DictOfSeries from :ref:`indexing <doc_indexing:Example dios>`.

        >>> di
            a |      b |      c |     d |
        ===== | ====== | ====== | ===== |
        0   0 | 2    5 | 4    7 | 6   0 |
        1   7 | 3    6 | 5   17 | 7   1 |
        2  14 | 4    7 | 6   27 | 8   2 |
        3  21 | 5    8 | 7   37 | 9   3 |
        4  28 | 6    9 | 8   47 | 10  4 |
        5  35 | 7   10 | 9   57 | 11  5 |
        6  42 | 8   11 | 10  67 | 12  6 |
        7  49 | 9   12 | 11  77 | 13  7 |
        8  56 | 10  13 | 12  87 | 14  8 |
        9  63 | 11  14 | 13  97 | 15  9 |

        >>> di.index_of()
        RangeIndex(start=0, stop=16, step=1)

        >>> di.index_of("shared")
        Int64Index([6, 7, 8, 9], dtype='int64')

        >>> di.index_of("uniques")
        Int64Index([0, 1, 14, 15], dtype='int64')
        """
        indexes = self.indexes
        if len(indexes) <= 1:
            return indexes.squeeze()

        if method in ["union", "all"]:
            res = ftools.reduce(pd.Index.union, indexes)
        elif method in ["intersection", "shared"]:
            res = ftools.reduce(pd.Index.intersection, indexes)
        elif method in ["uniques", "non-uniques"]:
            res = ftools.reduce(pd.Index.append, indexes)
            res = res.value_counts(sort=False, dropna=False)
            if method == "uniques":
                res = res[res == 1].index
            else:
                res = res[res > 1].index
        else:
            raise ValueError(method)
        return res if res.is_unique else res.unique()

    def squeeze(self, axis=None):
        """Squeeze a 1-dimensional axis objects into scalars."""
        if axis in [0, "index"]:
            if (self.lengths == 1).all():
                return self._data.apply(pd.Series.squeeze)
            return self
        elif axis in [1, "columns"]:
            if len(self) == 1:
                return self._data.squeeze()
            return self
        elif axis is None:
            if len(self) == 1:
                return self._data.squeeze().squeeze()
            if (self.lengths == 1).all():
                return self._data.apply(pd.Series.squeeze).squeeze()
            return self
        raise ValueError(axis)

    def dropna(self, inplace=False):
        """Return a bolean array that is `True` if the value is a Nan-value"""
        data = self.for_each("dropna", inplace=inplace)
        if inplace:
            return
        return self._construct_like_self(data=data, fastpath=True)

    def dropempty(self):
        """Drop empty columns. Return copy."""
        return self.loc[:, self.notempty()]

    def astype(self, dtype, copy=True, errors="raise"):
        """Cast the data to the given data type."""
        data = self.for_each("astype", dtype=dtype, copy=copy, errors=errors)
        return self._construct_like_self(data=data, fastpath=True)

    def _mask_or_where(self, cond, other=np.nan, inplace=False, mask=True):
        """helper to mask/where"""
        data = self if inplace else self.copy()

        if callable(other):
            other = other(data)

        if callable(cond):
            cond = cond(data)
        # if DictOfSeries is bool,
        # is already checked in aloc
        elif not _is_dios_like(cond):
            if not pdextra.is_bool_indexer(cond):
                raise ValueError(
                    "Object with boolean values only expected as condition"
                )

        if mask:
            data.aloc[cond] = other
        else:
            data.aloc[~cond] = other

        if inplace:
            return None
        return data

    def where(self, cond, other=np.nan, inplace=False):
        """
        Replace values where the condition is False.

        Parameters
        ----------
        cond : bool DictOfSeries, Series, array-like, or callable
            Where cond is True, keep the original value. Where False, replace
            with corresponding value from other. If cond is callable, it is computed
            on the DictOfSeries and should return boolean DictOfSeries or array.
            The callable must not change input DictOfSeries (though dios doesn’t check it).
            If cond is a bool Series, every column is (row-)aligned against it, before the
            boolean values are evaluated. Missing indices are treated like False values.

        other : scalar, Series, DictOfSeries, or callable
            Entries where cond is False are replaced with corresponding value from other.
            If other is callable, it is computed on the DictOfSeries and should return scalar
            or DictOfSeries. The callable must not change input DictOfSeries (though dios doesn’t check it).
            If other is a Series, every column is (row-)aligned against it, before the values
            are written. NAN's are written for missing indices.

        inplace : bool, default False
            Whether to perform the operation in place on the data.

        Returns
        -------
        DictOfSeries

        See Also
        --------
        mask: Mask data where condition is True
        """
        return self._mask_or_where(cond=cond, other=other, inplace=inplace, mask=False)

    def mask(self, cond, other=np.nan, inplace=False):
        """
        Replace values where the condition is True.

        Parameters
        ----------
        cond : bool DictOfSeries, Series, array-like, or callable
            Where cond is False, keep the original value. Where True, replace
            with corresponding value from other. If cond is callable, it is computed
            on the DictOfSeries and should return boolean DictOfSeries or array.
            The callable must not change input DictOfSeries (though dios doesn’t check it).
            If cond is a bool Series, every column is (row-)aligned against it, before the
            boolean values are evaluated. Missing indices are treated like False values.

        other : scalar, Series, DictOfSeries, or callable
            Entries where cond is True are replaced with corresponding value from other.
            If other is callable, it is computed on the DictOfSeries and should return scalar
            or DictOfSeries. The callable must not change input DictOfSeries (though dios doesn’t check it).
            If other is a Series, every column is (row-)aligned against it, before the values
            are written. NAN's are written for missing indices.

        inplace : bool, default False
            Whether to perform the operation in place on the data.

        Returns
        -------
        DictOfSeries

        See Also
        --------
        mask: Mask data where condition is False
        """
        return self._mask_or_where(cond=cond, other=other, inplace=inplace, mask=True)

    def memory_usage(self, index=True, deep=False):
        return self.for_each(pd.Series.memory_usage, index=index, deep=deep).sum()

    def to_df(self, how="outer"):
        """
        Transform DictOfSeries to a pandas.DataFrame.

        Because a pandas.DataFrame can not handle Series of different
        length, but DictOfSeries can, the missing data is filled with
        NaNs or is dropped, depending on the keyword `how`.

        Parameters
        ----------
        how: {'outer', 'inner'}, default 'outer'
            define how the resulting DataFrame index is generated:
            * 'outer': The indices of all columns, merged into one index is used.
                If a column misses values at the new index location, `NaN`s are filled.
            * 'inner': Only indices that are present in all columns are used, filling
                logic is not needed, but values are dropped, if a column has indices
                that are not known to all other columns.

        Returns
        -------
        pandas.DataFrame: transformed data

        Examples
        --------

        Missing data locations are filled with NaN's

        >>> a = pd.Series(11, index=range(2))
        >>> b = pd.Series(22, index=range(3))
        >>> c = pd.Series(33, index=range(1,9,3))
        >>> di = DictOfSeries(dict(a=a, b=b, c=c))
        >>> di
            a |     b |     c |
        ===== | ===== | ===== |
        0  11 | 0  22 | 1  33 |
        1  11 | 1  22 | 4  33 |
              | 2  22 | 7  33 |

        >>> di.to_df()
        columns     a     b     c
        0        11.0  22.0   NaN
        1        11.0  22.0  33.0
        2         NaN  22.0   NaN
        4         NaN   NaN  33.0
        7         NaN   NaN  33.0

        or is dropped if `how='inner'`

        >>> di.to_df(how='inner')
        columns   a   b   c
        1        11  22  33
        """
        if how == "inner":
            how = "shared"
        elif how == "outer":
            how = "all"
        else:
            raise ValueError(how)

        index = self.index_of(how)
        df = pd.DataFrame(columns=self.columns, index=index)
        for c, series in self.items():
            # this automatically respects the df-index, that
            # was set before. Missing locations are already
            # nans, present locations are set.
            df[c] = series.copy()

        df.attrs = self.attrs
        return df

    @property
    def debugDf(self):
        """Alias for ``to_df()`` as property, for debugging purpose."""
        return self.to_df()

    def min(self, axis=0, skipna=True):
        if axis is None:
            return self.for_each(pd.Series.min, skipna=skipna).min()
        if axis in [0, "index"]:
            return self.for_each(pd.Series.min, skipna=skipna)
        if axis in [1, "columns"]:
            func = lambda s1, s2: s1.where(s1 < s2, s2)
            return self.reduce_columns(func, skipna=skipna)
        raise ValueError(axis)

    def max(self, axis=0, skipna=None):
        if axis is None:
            return self.for_each(pd.Series.max, skipna=skipna).max()
        if axis in [0, "index"]:
            return self.for_each(pd.Series.max, skipna=skipna)
        if axis in [1, "columns"]:
            func = lambda s1, s2: s1.where(s1 > s2, s2)
            return self.reduce_columns(func, skipna=skipna)
        raise ValueError(axis)

    # ----------------------------------------------------------------------
    # Boolean and empty stuff

    def equals(self, other):
        """
        Test whether two DictOfSeries contain the same elements.

        This function allows two DictOfSeries to be compared against each other to see
        if they have the same shape and elements. NaNs in the same location are considered equal.
        The column headers do not need to have the same type, but the elements within the columns
        must be the same dtype.

        Parameters
        ----------
        other: DictOfSeries
            The other DictOfSeries to compare with.

        Returns
        -------
        bool
            True if all elements are the same in both DictOfSeries, False otherwise.
        """
        if not isinstance(other, _DiosBase):
            return False
        try:
            eq_nans = (self.isna() == other.isna()).all(None)
            eq_data = (self.dropna() == other.dropna()).all(None)
            eq_dtypes = (self.dtypes == other.dtypes).all()
            return eq_nans and eq_dtypes and eq_data
        except Exception:
            return False

    def isin(self, values):
        """Return a boolean dios, that indicates if the corresponding value is in the given array-like."""
        data = self.for_each("isin", values=values)
        return self._construct_like_self(data=data, fastpath=True)

    def all(self, axis=0):
        """
        Return whether all elements are True, potentially over an axis.

        Returns True unless there at least one element within a series
        or along a DictOfSeries axis that is False or equivalent (e.g. zero or empty).

        Parameters
        ----------
        axis : {0 or ‘index’, 1 or ‘columns’, None}, default 0
            Indicate which axis or axes should be reduced.
             * 0 / ‘index’ : reduce the index, return a Series whose index is the original column labels.
             * 1 / ‘columns’ : reduce the columns, return a Series whose index is the union of all columns indexes.
             * None : reduce all axes, return a scalar.

        Returns
        -------
        pandas.Series

        See Also
        --------
        pandas.Series.all: Return True if all elements are True.
        any: Return True if one (or more) elements are True.
        """
        if axis is None:
            return self._data.apply(all).all()
        if axis in [0, "index"]:
            return self._data.apply(all)
        if axis in [1, "columns"]:
            func = lambda s1, s2: s1.astype(bool) & s2.astype(bool)
            init = pd.Series(True, dtype=bool, index=self.index_of("all"))
            return self.reduce_columns(func, init)
        raise ValueError(axis)

    def any(self, axis=0):
        """
        Return whether any element is True, potentially over an axis.

        Returns False unless there at least one element within a series
        or along a DictOfSeries axis that is True or equivalent (e.g. non-zero or non-empty).

        Parameters
        ----------
        axis : {0 or ‘index’, 1 or ‘columns’, None}, default 0
            Indicate which axis or axes should be reduced.
             * 0 / ‘index’ : reduce the index, return a Series whose index is the original column labels.
             * 1 / ‘columns’ : reduce the columns, return a Series whose index is the union of all columns indexes.
             * None : reduce all axes, return a scalar.

        Returns
        -------
        pandas.Series

        See Also
        --------
        pandas.Series.any: Return whether any element is True.
        all: Return True if all elements are True.
        """
        if axis is None:
            return self._data.apply(any).any()
        if axis in [0, "index"]:
            return self._data.apply(any)
        if axis in [1, "columns"]:
            func = lambda s1, s2: s1.astype(bool) | s2.astype(bool)
            init = pd.Series(False, dtype=bool, index=self.index_of("all"))
            return self.reduce_columns(func, init)
        raise ValueError(axis)

    def isna(self, drop_empty=False):
        """
        Return a boolean DictOfSeries which indicates NA positions.
        """
        data = self.dropempty() if drop_empty else self
        data = data.for_each("isna")
        return self._construct_like_self(data=data, fastpath=True)

    def notna(self, drop_empty=False):
        """
        Return a boolean DictOfSeries which indicates non-NA positions.
        """
        data = self.dropempty() if drop_empty else self
        data = data.for_each("notna")
        return self._construct_like_self(data=data, fastpath=True)

    def hasnans(self, axis=0, drop_empty=False):
        """
        Returns a boolean Series along an axis, which indicates if it contains NA-entries.
        """
        data = self.dropempty() if drop_empty else self
        if axis is None:
            return data.for_each("hasnans").any()
        if axis in [0, "index"]:
            return data.for_each("hasnans")
        if axis in [1, "columns"]:
            func = lambda s1, s2: s1.isna() | s2.isna()
            init = pd.Series(False, dtype=bool, index=self.index_of("all"))
            return data.reduce_columns(func, init)
        raise ValueError(axis)

    def fillna(
        self,
        value=None,
        method=None,
        axis=None,
        inplace=False,
        limit=None,
        downcast=None,
    ):
        if axis in [None, 0, "index"]:
            kws = dict(value=value, method=method, limit=limit, downcast=downcast)
            data = self.for_each("fillna", inplace=inplace, **kws)
            if inplace:
                return
            return self._construct_like_self(data=data, fastpath=True)

        if axis in [1, "columns"]:
            raise NotImplementedError
        raise ValueError(axis)

    def isempty(self):
        """Returns a boolean Series, which indicates if an column is empty"""
        return self.for_each("empty").astype(bool)

    def notempty(self):
        """Returns a boolean Series, which indicates if an column is not empty"""
        return ~self.isempty()

    def isdata(self):
        """Alias for ``notna(drop_empty=True)``."""
        return self.notna(drop_empty=True)

    def isnull(self, drop_empty=False):
        """Alias for ``isna()``"""
        return self.isna(drop_empty=drop_empty)

    def notnull(self, drop_empty=False):
        """Alias, see ``notna()``."""
        return self.notna(drop_empty=drop_empty)

    def to_dios(self):
        """
        A dummy to allow unconditional to_dios calls
        on pd.DataFrame, pd.Series and dios.DictOfSeries
        """
        return self

    # ----------------------------------------------------------------------
    # Rendering Methods

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        repr = dios_options[OptsFields.dios_repr]
        showdim = self.lengths.max() > dios_options[OptsFields.disp_max_rows]
        return self.to_string(method=repr, show_dimensions=showdim)

    def to_string(
        self,
        max_rows=None,
        min_rows=None,
        max_cols=None,
        na_rep="NaN",
        show_dimensions=False,
        method=Opts.repr_indexed,
        no_value=" ",
        empty_series_rep="no data",
        col_delim=" | ",
        header_delim="=",
        col_space=None,
    ):
        """Pretty print a dios.

        if `method` == `indexed` (default):
            every column is represented by a own index and corresponding values

        if `method` == `aligned` [2]:
            one(!) global index is generated and values from a column appear at
            the corresponding index-location.

        Parameters
        ---------

        max_cols :
            not more column than `max_cols` are printed [1]

        max_rows :
            see `min_rows` [1]

        min_rows :
            not more rows than `min_rows` are printed, if rows of any series exceed `max_rows` [1]

        na_rep :
            all NaN-values are replaced by `na_rep`. Default `NaN`

        empty_series_rep :
            Ignored if not `method='indexed'`.
            Empty series are represented by the string in `empty_series_rep`

        col_delim : str
            Ignored if not `method='indexed'`.
            between all columns `col_delim` is inserted.

        header_delim :
            Ignored if not `method='indexed'`.
            between the column names (header) and the data, `header_delim` is inserted,
            if not None. The string is repeated, up to the width of the column. (str or None).

        no_value :
            Ignored if not `method='aligned'`.
            value that indicates, that no entry in the underling series is present. Bear in mind
            that this should differ from `na_rep`, otherwise you cannot differ missing- from NaN- values.

        Notes
        -----
            [1]: defaults to the corresponding value in `dios_options`
            [2]: the common-params are directly passed to pd.DataFrame.to_string(..)
            under the hood, if method is `aligned`

        """
        if self.empty:
            return _empty_repr(self)

        max_cols = max_cols or dios_options[OptsFields.disp_max_cols] or 100
        max_rows = max_rows or dios_options[OptsFields.disp_max_rows] or 200
        min_rows = min_rows or dios_options[OptsFields.disp_min_rows] or 100

        kwargs = dict(
            max_rows=max_rows,
            min_rows=min_rows,
            max_cols=max_cols,
            na_rep=na_rep,
            col_space=col_space,
            show_dimensions=show_dimensions,
        )

        if method == Opts.repr_aligned:
            return _to_aligned_df(self, no_value=no_value).to_string(**kwargs)

        # add pprint relevant options
        kwargs.update(
            empty_series_rep=empty_series_rep,
            col_delim=col_delim,
            header_delim=header_delim,
        )

        return pprint_dios(self, **kwargs)

    def to_csv(self, *args, **kwargs):
        self.to_df().to_csv(*args, **kwargs)

    to_csv.__doc__ = pd.DataFrame.to_csv.__doc__


def _empty_repr(di):
    return f"Empty DictOfSeries\n" f"Columns: {di.columns.to_list()}"


def pprint_dios(
    dios,
    max_rows=None,
    min_rows=None,
    max_cols=None,
    na_rep="NaN",
    empty_series_rep="no data",
    col_space=None,
    show_dimensions=True,
    col_delim=" | ",
    header_delim="=",
):
    na_rep = str(na_rep)
    empty_series_rep = str(empty_series_rep)
    col_delim = col_delim or " "

    min_rows = min(max_rows, min_rows)

    if dios.empty:
        return _empty_repr(dios)

    maxlen = dios.lengths.max()
    data = dios._data

    trunc_cols = len(data) > max_cols
    if trunc_cols:
        left, right = data.head(max_cols // 2), data.tail(max_cols // 2)
        data = left.append(right)

    # now data only contains series that we want to print.

    # if any series exceed max_rows we trim all series to min_rows
    series_lengths = data.apply(len).to_list()
    series_maxlen = max(series_lengths)
    trunc_rows = series_maxlen > max_rows
    max_rows = min_rows if trunc_rows else series_maxlen

    # we make a list of list, where the inner contains all
    # stringified values of the series upto max_rows+1, where
    # the additional row is the column-name
    outer = []
    for colname in data.index:
        s = data.at[colname]

        isempty = s.empty
        if isempty:
            s = pd.Series(empty_series_rep)
            idx = False
            cspace = col_space
        else:
            idx = True
            cspace = col_space // 2 if col_space else col_space

        sstr = s.to_frame().to_string(
            col_space=cspace,
            header=[str(colname)],
            index=idx,
            na_rep=na_rep,
            max_rows=max_rows,
            min_rows=min_rows,
        )
        li = sstr.split("\n")

        # HACK: empty series produce a unnecessary space,
        # because index is omitted
        if isempty:
            cstr, vstr = li
            if len(cstr.lstrip()) < len(vstr) and (cspace or 0) < len(vstr):
                li = [cstr[1:], vstr[1:]]

        outer.append(li)

    # now the length of every value-string per series are the same.
    # we need this length's to know, how many chars we need to fill,
    # once we exceed the length of the series, or if we insert whole
    # columns.
    valstr_len = [len(c[0]) for c in outer]

    rows = max_rows + 1  # colnames aka. header
    rows += 1 if trunc_rows else 0  # `...` in rows
    rows += 1 if header_delim else 0  # underline header

    if header_delim:
        for i, c in enumerate(outer):
            colheader = (header_delim * valstr_len[i])[: valstr_len[i]]
            c.insert(1, colheader)

    dots = " ... "
    if trunc_cols:
        outer.insert(max_cols // 2, [dots] * rows)
        valstr_len.insert(max_cols // 2, len(dots))
        series_lengths.insert(max_cols // 2, rows)

    txt = ""
    for r in range(rows):
        for i, c in enumerate(outer):
            try:
                vstr = c[r]
            except IndexError:
                vstr = " " * valstr_len[i]
            txt += vstr + col_delim
        txt += "\n"

    # add footer
    if show_dimensions:
        for i, c in enumerate(outer):
            # ignore the dot-column
            if trunc_cols and i == max_cols // 2:
                txt += dots + " " * len(col_delim)
            else:
                txt += f"[{series_lengths[i]}]".ljust(valstr_len[i] + len(col_delim))

        txt += f"\n\nmax: [{maxlen} rows x {len(dios.columns)} columns]"
        txt += "\n"

    return txt


def _to_aligned_df(dios, no_value=" "):
    if dios.empty:
        return pd.DataFrame(columns=dios.columns)

    # keep track of all real nans
    nandict = {}
    for c in dios:
        nans = dios[c].isna()
        nandict[c] = nans[nans].index

    df = dios.to_df()
    df[df.isna()] = no_value

    # reinsert all real nans
    for c in df:
        df.loc[nandict[c], c] = np.nan

    return df


def to_dios(obj) -> DictOfSeries:
    if isinstance(obj, DictOfSeries):
        return obj
    return DictOfSeries(data=obj)


def __monkey_patch_pandas():
    def to_dios(self):
        return DictOfSeries(data=self)

    pd.Series.to_dios = to_dios
    pd.DataFrame.to_dios = to_dios


__monkey_patch_pandas()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from saqc.core.register import register, flagging, processing
from saqc.core.flags import Flags, initFlagsLike
from saqc.core.core import SaQC
from saqc.core.translation import (
    FloatScheme,
    DmpScheme,
    PositionalScheme,
    SimpleScheme,
)
from saqc.core.reader import fromConfig



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Optional, Callable

import numpy as np
import pandas as pd
from typing_extensions import Literal

from saqc.constants import BAD
import saqc
from saqc.funcs.interpolation import _SUPPORTED_METHODS


class Resampling:
    def linear(
        self,
        field: str,
        freq: str,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("linear", locals())

    def interpolate(
        self,
        field: str,
        freq: str,
        method: _SUPPORTED_METHODS,
        order: int = 1,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("interpolate", locals())

    def shift(
        self,
        field: str,
        freq: str,
        method: Literal["fshift", "bshift", "nshift"] = "nshift",
        freq_check: Optional[Literal["check", "auto"]] = None,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("shift", locals())

    def resample(
        self,
        field: str,
        freq: str,
        func: Callable[[pd.Series], pd.Series] = np.mean,
        method: Literal["fagg", "bagg", "nagg"] = "bagg",
        maxna: Optional[int] = None,
        maxna_group: Optional[int] = None,
        maxna_flags: Optional[int] = None,  # TODO: still a case ??
        maxna_group_flags: Optional[int] = None,
        flag_func: Callable[[pd.Series], float] = max,
        freq_check: Optional[Literal["check", "auto"]] = None,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("resample", locals())

    def concatFlags(
        self,
        field: str,
        target: str,
        method: Literal[
            "inverse_fagg",
            "inverse_bagg",
            "inverse_nagg",
            "inverse_fshift",
            "inverse_bshift",
            "inverse_nshift",
            "inverse_interpolation",
        ],
        freq: Optional[str] = None,
        drop: Optional[bool] = False,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("concatFlags", locals())



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Any, Union

import pandas as pd
import numpy as np
from dios import DictOfSeries
from typing_extensions import Literal

from saqc.constants import BAD
import saqc


class FlagTools:
    def clearFlags(self, field: str, **kwargs) -> saqc.SaQC:
        return self._defer("clearFlags", locals())

    def forceFlags(self, field: str, flag: float = BAD, **kwargs) -> saqc.SaQC:
        return self._defer("forceFlags", locals())

    def flagDummy(self, field: str, **kwargs) -> saqc.SaQC:
        return self._defer("flagDummy", locals())

    def flagUnflagged(self, field: str, flag: float = BAD, **kwargs) -> saqc.SaQC:
        return self._defer("flagUnflagged", locals())

    def flagManual(
        self,
        field: str,
        mdata: Union[pd.Series, pd.DataFrame, DictOfSeries, list, np.array],
        method: Literal[
            "left-open", "right-open", "closed", "plain", "ontime"
        ] = "left-open",
        mformat: Literal["start-end", "mflag"] = "start-end",
        mflag: Any = 1,
        flag: float = BAD,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagManual", locals())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# imports needed to make the functions register themself
from saqc.core.register import register
from saqc.funcs.breaks import *
from saqc.funcs.changepoints import *
from saqc.funcs.constants import *
from saqc.funcs.curvefit import *
from saqc.funcs.drift import *
from saqc.funcs.generic import *
from saqc.funcs.interpolation import *
from saqc.funcs.outliers import *
from saqc.funcs.pattern import *
from saqc.funcs.resampling import *
from saqc.funcs.residues import *
from saqc.funcs.rolling import *
from saqc.funcs.scores import *
from saqc.funcs.tools import *
from saqc.funcs.transformation import *
from saqc.funcs.flagtools import *
from saqc.funcs.noise import *



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Sequence, Tuple, Union

import numpy as np
import pandas as pd

from dios import DictOfSeries

from saqc.constants import BAD, UNFLAGGED, ENVIRONMENT, FILTER_ALL
from saqc.core.history import History
from saqc.lib.tools import toSequence
from saqc.lib.types import GenericFunction, PandasLike
from saqc.core.flags import Flags
from saqc.core.register import register, _isflagged, FunctionWrapper


def _prepare(
    data: DictOfSeries, flags: Flags, columns: Sequence[str], dfilter: float
) -> Tuple[DictOfSeries, Flags]:
    fchunk = Flags({f: flags[f] for f in columns})
    dchunk, _ = FunctionWrapper._maskData(
        data=data.loc[:, columns].copy(), flags=fchunk, columns=columns, thresh=dfilter
    )
    return dchunk, fchunk.copy()


def _execGeneric(
    flags: Flags,
    data: PandasLike,
    func: GenericFunction,
    dfilter: float = FILTER_ALL,
) -> DictOfSeries:

    globs = {
        "isflagged": lambda data: _isflagged(flags[data.name], thresh=dfilter),
        **ENVIRONMENT,
    }

    func.__globals__.update(globs)

    if isinstance(data, pd.Series):
        data = data.to_frame()

    out = func(*[data[c] for c in data.columns])
    return DictOfSeries(out)


@register(
    mask=[],
    demask=[],
    squeeze=[],
    multivariate=True,
    handles_target=True,
)
def processGeneric(
    data: DictOfSeries,
    field: str | Sequence[str],
    flags: Flags,
    func: GenericFunction,
    target: str | Sequence[str] = None,
    flag: float = UNFLAGGED,
    dfilter: float = FILTER_ALL,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Generate/process data with user defined functions.

    Formally, what the function does, is the following:

    1.  Let F be a Callable, depending on fields f_1, f_2,...f_K, (F = F(f_1, f_2,...f_K))
        Than, for every timestamp t_i that occurs in at least one of the timeseries data[f_j] (outer join),
        The value v_i is computed via:
        v_i = data([f_1][t_i], data[f_2][t_i], ..., data[f_K][t_i]), if all data[f_j][t_i] do exist
        v_i = ``np.nan``, if at least one of the data[f_j][t_i] is missing.
    2.  The result is stored to ``data[target]``, if ``target`` is given or to ``data[field]`` otherwise

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str or list of str
        The variable(s) passed to func.
    flags : saqc.Flags
        Container to store flags of the data.
    func : callable
        Function to call on the variables given in ``field``. The return value will be written
        to ``target`` or ``field`` if the former is not given. This implies, that the function
        needs to accept the same number of arguments (of type pandas.Series) as variables given
        in ``field`` and should return an iterable of array-like objects with the same number
        of elements as given in ``target`` (or ``field`` if ``target`` is not specified).
    target: str or list of str
        The variable(s) to write the result of ``func`` to. If not given, the variable(s)
        specified in ``field`` will be overwritten. If a ``target`` is not given, it will be
        created.
    flag: float, default ``UNFLAGGED``
        The quality flag to set. The default ``UNFLAGGED`` states the general idea, that
        ``processGeneric`` generates 'new' data without direct relation to the potentially
        already present flags.
    dfilter: float, default ``FILTER_ALL``
        Threshold flag. Flag values greater than ``dfilter`` indicate that the associated
        data value is inappropiate for further usage.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        The shape of the data may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
        The flags shape may have changed relatively to the input flags.

    Note
    -----
    All the numpy functions are available within the generic expressions.

    Examples
    --------
    Compute the sum of the variables 'rainfall' and 'snowfall' and save the result to
    a (new) variable 'precipitation'

    >>> saqc.genericProcess(field=["rainfall", "snowfall"], target="precipitation'", func=lambda x, y: x + y)
    """

    fields = toSequence(field)
    targets = fields if target is None else toSequence(target)

    dchunk, fchunk = _prepare(data, flags, fields, dfilter)
    result = _execGeneric(fchunk, dchunk, func, dfilter=dfilter)

    meta = {
        "func": "procGeneric",
        "args": (),
        "kwargs": {
            "field": field,
            "func": func.__name__,
            "target": target,
            "flag": flag,
            "dfilter": dfilter,
        },
    }

    # update data & flags
    for i, col in enumerate(targets):

        datacol = result.iloc[:, i]
        data[col] = datacol

        if col not in flags:
            flags.history[col] = History(datacol.index)

        if not flags[col].index.equals(datacol.index):
            raise ValueError(
                f"cannot assign function result to the existing variable {repr(col)} "
                "because of incompatible indices, please choose another 'target'"
            )

        flags.history[col].append(pd.Series(flag, index=datacol.index), meta)

    return data, flags


@register(
    mask=[],
    demask=[],
    squeeze=[],
    multivariate=True,
    handles_target=True,
)
def flagGeneric(
    data: DictOfSeries,
    field: Union[str, Sequence[str]],
    flags: Flags,
    func: GenericFunction,
    target: Union[str, Sequence[str]] = None,
    flag: float = BAD,
    dfilter: float = FILTER_ALL,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Flag data with user defined functions.

    Formally, what the function does, is the following:
    Let X be a Callable, depending on fields f_1, f_2,...f_K, (X = X(f_1, f_2,...f_K))
    Than for every timestamp t_i in data[field]:
    data[field][t_i] is flagged if X(data[f_1][t_i], data[f_2][t_i], ..., data[f_K][t_i]) is True.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str or list of str
        The variable(s) passed to func.
    flags : saqc.Flags
        Container to store flags of the data.
    func : callable
        Function to call on the variables given in ``field``. The function needs to accept the same
        number of arguments (of type pandas.Series) as variables given in ``field`` and return an
        iterable of array-like objects of with dtype bool and with the same number of elements as
        given in ``target`` (or ``field`` if ``target`` is not specified). The function output
        determines the values to flag.
    target: str or list of str
        The variable(s) to write the result of ``func`` to. If not given, the variable(s)
        specified in ``field`` will be overwritten. If a ``target`` is not given, it will be
        created.
    flag: float, default ``UNFLAGGED``
        The quality flag to set. The default ``UNFLAGGED`` states the general idea, that
        ``processGeneric`` generates 'new' data without direct relation to the potentially
        already present flags.
    dfilter: float, default ``FILTER_ALL``
        Threshold flag. Flag values greater than ``dfilter`` indicate that the associated
        data value is inappropiate for further usage.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values may have changed relatively to the flags input.

    Note
    -----
    All the numpy functions are available within the generic expressions.

    Examples
    --------

    1. Flag the variable 'rainfall', if the sum of the variables 'temperature' and 'uncertainty' is below zero:

    >>> saqc.flagGeneric(field=["temperature", "uncertainty"], target="rainfall", func= lambda x, y: temperature + uncertainty < 0

    2. Flag the variable 'temperature', where the variable 'fan' is flagged:

    >>> saqc.flagGeneric(field="fan", target="temperature", func=lambda x: isflagged(x))

    3. The generic functions also support all pandas and numpy functions:

    >>> saqc.flagGeneric(field="fan", target="temperature", func=lambda x: np.sqrt(x) < 7)
    """

    fields = toSequence(field)
    targets = fields if target is None else toSequence(target)

    dchunk, fchunk = _prepare(data, flags, fields, dfilter)
    result = _execGeneric(fchunk, dchunk, func, dfilter=dfilter)

    if len(targets) != len(result.columns):
        raise ValueError(
            f"the generic function returned {len(result.columns)} field(s), but only {len(targets)} target(s) were given"
        )

    if not (result.dtypes == bool).all():
        raise TypeError(f"generic expression does not return a boolean array")

    meta = {
        "func": "flagGeneric",
        "args": (),
        "kwargs": {
            "field": field,
            "func": func.__name__,
            "target": target,
            "flag": flag,
            "dfilter": dfilter,
        },
    }

    # update flags & data
    for i, col in enumerate(targets):

        maskcol = result.iloc[:, i]

        # make sure the column exists
        if col not in flags:
            flags.history[col] = History(maskcol.index)

        # dummy column to ensure consistency between flags and data
        if col not in data:
            data[col] = pd.Series(np.nan, index=maskcol.index)

        maskcol = maskcol & ~_isflagged(flags[col], dfilter)
        flagcol = maskcol.replace({False: np.nan, True: flag}).astype(float)

        # we need equal indices to work on
        if not flags[col].index.equals(maskcol.index):
            raise ValueError(
                f"cannot assign function result to the existing variable {repr(col)} "
                "because of incompatible indices, please choose another 'target'"
            )

        flags.history[col].append(flagcol, meta)

    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from saqc.constants import BAD, FILTER_ALL
import saqc


class Breaks:
    def flagMissing(
        self, field: str, flag: float = BAD, dfilter: float = FILTER_ALL, **kwargs
    ) -> saqc.SaQC:
        return self._defer("flagMissing", locals())

    def flagIsolated(
        self,
        field: str,
        gap_window: str,
        group_window: str,
        flag: float = BAD,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("flagIsolated", locals())

    def flagJumps(
        self,
        field: str,
        thresh: float,
        window: str,
        min_periods: int = 1,
        flag: float = BAD,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("flagJumps", locals())



from .test_setup import *
from pandas.core.dtypes.common import is_scalar


@pytest.mark.parametrize(
    ("idxer", "exp"),
    [(0, s1), (1, s2), (2, s3), (3, s4), (-1, s4), (-2, s3), (-3, s2), (-4, s1)],
)
def test__getitem_single_iloc(dios_aligned, idxer, exp):
    di = dios_aligned.iloc[:, idxer]
    assert isinstance(di, pd.Series)
    assert (di == exp).all()


@pytest.mark.parametrize(
    ("idxer", "exp"), [((1, 0), s1), ((3, -2), s3), ((-1, -1), s4)]
)
def test__getitem_scalar_iloc(dios_aligned, idxer, exp):
    di = dios_aligned.iloc[idxer]
    assert is_scalar(di)
    assert di == exp.iloc[idxer[0]]


@pytest.mark.parametrize(
    "idxer",
    [
        -5,
        99,
        "a",
        "2",
        None,
    ],
)
def test__getitem_single_iloc_fail(dios_aligned, idxer):
    with pytest.raises((KeyError, IndexError, TypeError)):
        di = dios_aligned.iloc[:, idxer]


# #############################
# __SETITEM__



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# see test/functs/fixtures.py for global fixtures "course_..."
import dios
from tests.fixtures import *

from saqc.funcs.outliers import (
    flagMAD,
    flagOffset,
    flagRaise,
    flagMVScores,
    flagByGrubbs,
)
from saqc.constants import *
from saqc.core import initFlagsLike


@pytest.fixture(scope="module")
def spiky_data():
    index = pd.date_range(start="2011-01-01", end="2011-01-05", freq="5min")
    s = pd.Series(np.linspace(1, 2, index.size), index=index, name="spiky_data")
    s.iloc[100] = 100
    s.iloc[1000] = -100
    flag_assertion = [100, 1000]
    return dios.DictOfSeries(s), flag_assertion


def test_flagMad(spiky_data):
    data = spiky_data[0]
    field, *_ = data.columns
    flags = initFlagsLike(data)
    data, flags_result = flagMAD(data, field, flags, "1H", flag=BAD)
    flag_result = flags_result[field]
    test_sum = (flag_result[spiky_data[1]] == BAD).sum()
    assert test_sum == len(spiky_data[1])


def test_flagSpikesBasic(spiky_data):
    data = spiky_data[0]
    field, *_ = data.columns
    flags = initFlagsLike(data)
    data, flags_result = flagOffset(
        data, field, flags, thresh=60, tolerance=10, window="20min", flag=BAD
    )
    flag_result = flags_result[field]
    test_sum = (flag_result[spiky_data[1]] == BAD).sum()
    assert test_sum == len(spiky_data[1])


# see test/functs/fixtures.py for the 'course_N'
@pytest.mark.parametrize(
    "dat",
    [
        pytest.lazy_fixture("course_1"),
        pytest.lazy_fixture("course_2"),
        pytest.lazy_fixture("course_3"),
        pytest.lazy_fixture("course_4"),
    ],
)
def test_flagSpikesLimitRaise(dat):
    data, characteristics = dat()
    field, *_ = data.columns
    flags = initFlagsLike(data)
    _, flags_result = flagRaise(
        data,
        field,
        flags,
        thresh=2,
        freq="10min",
        raise_window="20min",
        numba_boost=False,
        flag=BAD,
    )
    assert np.all(flags_result[field][characteristics["raise"]] > UNFLAGGED)
    assert not np.any(flags_result[field][characteristics["return"]] > UNFLAGGED)
    assert not np.any(flags_result[field][characteristics["drop"]] > UNFLAGGED)


# see test/functs/fixtures.py for the 'course_N'
@pytest.mark.parametrize("dat", [pytest.lazy_fixture("course_3")])
def test_flagMultivarScores(dat):
    data1, characteristics = dat(
        periods=1000, initial_level=5, final_level=15, out_val=50
    )
    data2, characteristics = dat(
        periods=1000, initial_level=20, final_level=1, out_val=30
    )
    fields = ["data1", "data2"]
    s1, s2 = data1.squeeze(), data2.squeeze()
    s1 = pd.Series(data=s1.values, index=s1.index)
    s2 = pd.Series(data=s2.values, index=s1.index)
    data = dios.DictOfSeries([s1, s2], columns=["data1", "data2"])
    flags = initFlagsLike(data)
    _, flags_result = flagMVScores(
        data=data,
        field=fields,
        flags=flags,
        trafo=np.log,
        iter_start=0.95,
        n=10,
        flag=BAD,
    )
    for field in fields:
        isflagged = flags_result[field] > UNFLAGGED
        assert isflagged[characteristics["raise"]].all()
        assert not isflagged[characteristics["return"]].any()
        assert not isflagged[characteristics["drop"]].any()


@pytest.mark.parametrize("dat", [pytest.lazy_fixture("course_3")])
def test_grubbs(dat):
    data, char_dict = dat(
        freq="10min",
        periods=45,
        initial_level=0,
        final_level=0,
        crowd_size=1,
        crowd_spacing=3,
        out_val=-10,
    )
    flags = initFlagsLike(data)
    data, result_flags = flagByGrubbs(
        data, "data", flags, window=20, min_periods=15, flag=BAD
    )
    assert np.all(result_flags["data"][char_dict["drop"]] > UNFLAGGED)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Callable, Tuple

import numpy as np
from typing_extensions import Literal

from saqc.constants import BAD
import saqc


class ChangePoints:
    def flagChangePoints(
        self,
        field: str,
        stat_func: Callable[[np.ndarray, np.ndarray], float],
        thresh_func: Callable[[np.ndarray, np.ndarray], float],
        window: str | Tuple[str, str],
        min_periods: int | Tuple[int, int],
        closed: Literal["right", "left", "both", "neither"] = "both",
        reduce_window: str = None,
        reduce_func: Callable[[np.ndarray, np.ndarray], int] = lambda x, _: x.argmax(),
        flag: float = BAD,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagChangePoints", locals())

    def assignChangePointCluster(
        self,
        field: str,
        stat_func: Callable[[np.array, np.array], float],
        thresh_func: Callable[[np.array, np.array], float],
        window: str | Tuple[str, str],
        min_periods: int | Tuple[int, int],
        closed: Literal["right", "left", "both", "neither"] = "both",
        reduce_window: str = None,
        reduce_func: Callable[
            [np.ndarray, np.ndarray], float
        ] = lambda x, _: x.argmax(),
        model_by_resids: bool = False,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("assignChangePointCluster", locals())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np

from saqc.constants import *
from saqc.funcs.constants import flagConstants, flagByVariance
from saqc.core import initFlagsLike, Flags

from tests.common import initData


@pytest.fixture
def data():
    constants_data = initData(
        1, start_date="2011-01-01 00:00:00", end_date="2011-01-01 03:00:00", freq="5min"
    )
    constants_data.iloc[5:25] = 200
    return constants_data


def test_constants_flagBasic(data):
    field, *_ = data.columns
    flags = initFlagsLike(data)
    data, flags_result = flagConstants(
        data, field, flags, window="15Min", thresh=0.1, flag=BAD
    )
    flagscol = flags_result[field]
    assert np.all(flagscol[5:25] == BAD)
    assert np.all(flagscol[:5] == UNFLAGGED)
    assert np.all(flagscol[25 + 1 :] == UNFLAGGED)


def test_constants_flagVarianceBased(data):
    expected = np.arange(5, 25)
    field, *_ = data.columns
    flags = initFlagsLike(data)
    data, flags_result1 = flagByVariance(data, field, flags, window="1h", flag=BAD)

    flagscol = flags_result1[field]
    assert np.all(flagscol[5:25] == BAD)
    assert np.all(flagscol[:5] == UNFLAGGED)
    assert np.all(flagscol[25 + 1 :] == UNFLAGGED)



#!/usr/bin/env python
from __future__ import annotations

from copy import deepcopy, copy as shallowcopy
import itertools

from typing import Dict, Tuple, Type, Union, List, Any
from typing_extensions import Literal
import pandas as pd
import numpy as np

from saqc.constants import *


class History:
    """
    Saqc internal storage for the history of a (single) flags column.

    The flag-history (FH) stores the history of a flags column. Each time
    ``append`` is called a new column is appended to the FH. The column
    names are increasing integers starting with 0. After initialisation
    the FH is empty and has no columns at all.

    To get the worst flags (highest value) that are currently stored in
    the FH, we provide a ``max()`` method. It returns a pd.Series indicating
    the worst flag per row.

    For more details and a detailed discussion, why this is needed, how this
    works and possible other implementations, see #GL143 [1].

    [1] https://git.ufz.de/rdm-software/saqc/-/issues/143

    Parameters
    ----------
    index: pd.Index
        A index that fit the flags to be insert.

    See Also
    --------
    createHistoryFromData: function to create History from existing data
    """

    def __init__(self, index: pd.Index):

        self.hist = pd.DataFrame(index=index)
        self.meta = []

    @property
    def index(self) -> pd.Index:
        """
        The index of FH.

        The index is the same for all columns.

        Notes
        -----
        The index should always be equal to the flags series,
        the FH is associated with. If this is messed up
        something went wrong in saqc internals or in a user-
        defined test.

        Returns
        -------
        index : pd.Index
        """
        return self.hist.index

    @property
    def columns(self) -> pd.Index:
        """
        Columns of the FH.

        The columns are always continuously
        increasing integers, starting from 0.

        Returns
        -------
        columns : pd.Index
        """
        return self.hist.columns

    @property
    def empty(self) -> bool:
        """
        Indicator whether History is empty.

        True if History is entirely empty (no items).

        Returns
        -------
        bool
            If History is empty, return True, if not return False.
        """
        return len(self) == 0

    def _insert(self, s: pd.Series, pos: int) -> History:
        """
        Insert data at an arbitrary position in the FH.

        No validation of series is done here.

        Parameters
        ----------
        s : pd.Series
            the series to insert

        pos : int
            the position to insert

        Returns
        -------
        History
        """
        # Note:
        # all following code must handle a passed empty series

        # ensure continuous increasing columns
        assert 0 <= pos <= len(self)

        self.hist[pos] = s.astype("category")

        return self

    def append(self, value: Union[pd.Series, History], meta: dict = None) -> History:
        """
        Create a new FH column and insert given pd.Series to it.

        Parameters
        ----------
        value : pd.Series or History
            The data to append. Must have dtype float and the index must
            match the index of the History.

        meta : dict, default None
            metadata dictionary to store with the series. Ignored if ``value`` is of
             type History. None defaults to a empty dictionary.

        Returns
        -------
        history with appended series

        Raises
        ------
        TypeError: if value is not pd.Series
        ValueError: on index miss-match or wrong dtype
        """
        if isinstance(value, History):
            return self._appendHistory(value)

        if not isinstance(value, pd.Series):
            raise TypeError("'value' is not a pd.Series")

        if meta is None:
            meta = {}

        if not isinstance(meta, dict):
            raise TypeError("'meta' must be of type None or dict")

        val = self._validateValue(value)
        if not val.index.equals(self.index):
            raise ValueError("Index does not match")

        self._insert(val, pos=len(self))
        self.meta.append(meta.copy())
        return self

    def _appendHistory(self, value: History):
        """
        Append multiple columns of a history to self.

        Parameters
        ----------
        value : History
            Holding the columns to append

        Returns
        -------
        History with appended columns.

        Raises
        ------
        ValueError : If the index of the passed history does not match.

        Notes
        -----
        This ignores the column names of the passed History.
        """
        self._validate(value.hist, value.meta)
        if not value.index.equals(self.index):
            raise ValueError("Index does not match")

        # we copy shallow because we only want to set new columns
        # the actual data copy happens in calls to astype
        value_hist = value.hist.copy(deep=False)
        value_meta = value.meta.copy()

        # rename columns, to avoid ``pd.DataFrame.loc`` become confused
        n = len(self.columns)
        columns = pd.Index(range(n, n + len(value_hist.columns)))
        value_hist.columns = columns

        hist = self.hist.astype(float)
        hist.loc[:, columns] = value_hist.astype(float)
        self.hist = hist.astype("category")
        self.meta += value_meta
        return self

    def max(self, raw=False) -> pd.Series:
        """
        Get the maximum value per row of the FH.

        Returns
        -------
        pd.Series: maximum values
        """
        result = self.hist.astype(float)
        if result.empty:
            result = pd.DataFrame(data=np.nan, index=self.hist.index, columns=[0])

        result = result.ffill(axis=1).iloc[:, -1]

        if raw:
            return result
        else:
            return result.fillna(UNFLAGGED)

    def reindex(
        self, index: pd.Index, fill_value_last: float = UNFLAGGED, copy: bool = True
    ) -> History:
        """
        Reindex the History. Be careful this alters the past.

        Parameters
        ----------
        index : pd.Index
            the index to reindex to.

        fill_value_last : float, default UNFLAGGED
            value to fill nan's in the last column.
            Defaults to 0 (UNFLAGGED).

        copy : bool, default True
            If False, alter the underlying history, otherwise return a copy.

        Returns
        -------
        History
        """
        out = self.copy() if copy else self

        hist = out.hist.astype(float).reindex(
            index=index, copy=False, fill_value=np.nan
        )

        # Note: all following code must handle empty frames
        hist.iloc[:, -1:] = hist.iloc[:, -1:].fillna(fill_value_last)

        out.hist = hist.astype("category")

        return out

    def apply(
        self,
        index: pd.Index,
        func: callable,
        func_kws: dict,
        func_handle_df: bool = False,
        copy: bool = True,
    ):
        """
        Apply a function on each column in history.

        The function must take a `pd.Series` as first arg, which is a column from
        `hist`. If ``func_handle_df=True`` each functions take a ``pd.DataFrame``
        as first argument, holding all columns at once.
        Bear in mind:
        - the functions mustn't alter the passed objects
        - the functions are not allowed to add or remove columns
        - the function must return same type as first argument
        - the returned object must have same index as the passed ``index`` to ``apply``
            as first argument

        Parameters
        ----------
        index: pd.Index
            Index the new history should have. This is used to ensure the passed
            functions worked correct and also used if the function does not apply,
            because the initial history is empty. Then the altered empty history is
            reindexed to this index.

        func : callable
            function to apply on `History.hist` (flags DataFrame)

        func_kws : dict
            hist-function keywords dict

        func_handle_df : bool, default False
            If `True`, the Dataframe under `History`.hist is passed to the given functions,
            thus the function must handle `pd.Dataframes` as first input. If `False`, each
            column is passed separately, thus the function must handle those.

        copy : bool, default True
            If False, alter the underlying history, otherwise return a copy.


        Returns
        -------
        history with altered columns
        """
        hist = pd.DataFrame(index=index)

        # implicit copy by astype
        # convert data to floats as functions may fail with categoricals
        if func_handle_df:
            hist = func(self.hist.astype(float), **func_kws)
        else:
            for pos in self.columns:
                hist[pos] = func(self.hist[pos].astype(float), **func_kws)

        History._validate(hist, self.meta)

        if copy:
            history = History(index=None)  # noqa
            history.meta = self.meta.copy()
        else:
            history = self

        history.hist = hist.astype("category")

        return history

    def copy(self, deep=True) -> History:
        """
        Make a copy of the FH.

        Parameters
        ----------
        deep : bool, default True
            - ``True``: make a deep copy
            - ``False``: make a shallow copy

        Returns
        -------
        copy : History
            the copied FH
        """
        copyfunc = deepcopy if deep else shallowcopy
        new = History(self.index)
        new.hist = self.hist.copy(deep)
        new.meta = copyfunc(self.meta)
        return new

    def __copy__(self):
        return self.copy(deep=False)

    def __deepcopy__(self, memo=None):
        """
        Parameters
        ----------
        memo, default None
            Standard signature. Unused
        """
        return self.copy(deep=True)

    def __len__(self) -> int:
        return len(self.hist.columns)

    def __repr__(self):

        if self.empty:
            return str(self.hist).replace("DataFrame", "History")

        r = self.hist.astype(str)

        return str(r)[1:]

    # --------------------------------------------------------------------------------
    # validation
    #

    @staticmethod
    def _validate(hist: pd.DataFrame, meta: List[Any]) -> Tuple[pd.DataFrame, List]:
        """
        check type, columns, index, dtype of hist and if the meta fits also
        """

        # check hist
        if not isinstance(hist, pd.DataFrame):
            raise TypeError(
                f"'hist' must be of type pd.DataFrame, but is of type {type(hist).__name__}"
            )
        # isin([float, ..]) does not work !
        if not (
            (hist.dtypes == float)
            | (hist.dtypes == np.float32)
            | (hist.dtypes == np.float64)
            | (hist.dtypes == "category")
        ).all():
            raise ValueError(
                "dtype of all columns in hist must be float or categorical"
            )

        if not hist.empty and (
            not hist.columns.equals(pd.Index(range(len(hist.columns))))
            or not np.issubdtype(hist.columns.dtype, np.integer)
        ):
            raise ValueError(
                "column names must be continuous increasing int's, starting with 0."
            )

        # check meta
        if not isinstance(meta, list):
            raise TypeError(
                f"'meta' must be of type list, but is of type {type(meta).__name__}"
            )
        if not all([isinstance(e, dict) for e in meta]):
            raise TypeError("All elements in meta must be of type 'dict'")

        # check combinations of hist and meta
        if not len(hist.columns) == len(meta):
            raise ValueError(
                "'meta' must have as many entries as columns exist in hist"
            )

        return hist, meta

    @staticmethod
    def _validateValue(obj: pd.Series) -> pd.Series:
        """
        index is not checked !
        """
        if not isinstance(obj, pd.Series):
            raise TypeError(
                f"value must be of type pd.Series, but {type(obj).__name__} was given"
            )

        if not ((obj.dtype == float) or isinstance(obj.dtype, pd.CategoricalDtype)):
            raise ValueError("dtype must be float or categorical")

        return obj


def createHistoryFromData(
    hist: pd.DataFrame,
    meta: List[Dict],
    copy: bool = False,
):
    """
    Create a History from existing data.

    Parameters
    ----------
    hist : pd.Dataframe
        Data that define the flags of the history.

    meta : List of dict
        A list holding meta information for each column, therefore it must
        have the same number of entries as columns exist in `hist`.

    copy : bool, default False
        If `True`, the input data is copied, otherwise not.


    Notes
    -----
    To create a very simple History from a flags dataframe ``f`` use
    ``mask = pd.DataFrame(True, index=f.index, columns=f.columns``
    and
    ``meta = [{}] * len(f.columns)``.

    Returns
    -------
    History
    """
    History._validate(hist, meta)

    if copy:
        hist = hist.copy()
        meta = deepcopy(meta)

    history = History(index=None)  # noqa
    history.hist = hist.astype("category", copy=False)
    history.meta = meta
    return history



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Optional, Union, Callable

import numpy as np
import pandas as pd
from typing_extensions import Literal

from saqc.constants import BAD
import saqc


class Residues:
    def calculatePolynomialResidues(
        self,
        field: str,
        window: Union[str, int],
        order: int,
        min_periods: Optional[int] = 0,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("calculatePolynomialResidues", locals())

    def calculateRollingResidues(
        self,
        field: str,
        window: Union[str, int],
        func: Callable[[pd.Series], np.ndarray] = np.mean,
        min_periods: Optional[int] = 0,
        center: bool = True,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("calculateRollingResidues", locals())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import io
import ast
from pathlib import Path
import pandas as pd

from saqc.core.core import SaQC
from saqc.core.visitor import ConfigFunctionParser
from saqc.lib.tools import isQuoted


COMMENT = "#"
SEPARATOR = ";"


def readFile(fname) -> pd.DataFrame:

    fobj = (
        io.open(fname, "r", encoding="utf-8")
        if isinstance(fname, (str, Path))
        else fname
    )

    out = []
    for i, line in enumerate(fobj):
        row = line.strip().split(COMMENT, 1)[0]
        if not row:
            # skip over comment line
            continue

        parts = [p.strip() for p in row.split(SEPARATOR)]
        if len(parts) != 2:
            raise RuntimeError(
                "The configuration format expects exactly two columns, one "
                "for the variable name and one for the test to apply, but "
                f"in line {i} we got: \n'{line}'"
            )
        out.append([i + 1] + parts)

    try:
        fobj.close()
    except AttributeError:
        pass

    df = pd.DataFrame(
        out[1:],
        columns=[
            "row",
        ]
        + out[0][1:],
    ).set_index("row")
    return df


def fromConfig(fname, *args, **kwargs):
    saqc = SaQC(*args, **kwargs)
    config = readFile(fname)

    for _, field, expr in config.itertuples():

        regex = False
        if isQuoted(field):
            fld = field[1:-1]
            regex = True
        else:
            fld = field

        try:
            tree = ast.parse(expr, mode="eval")
            func_name, kwargs = ConfigFunctionParser().parse(tree.body)
        except Exception as e:
            raise type(e)(f"failed to parse: {field} ; {expr}") from e

        kwargs["field" if "field" not in kwargs else "target"] = fld
        try:
            saqc = getattr(saqc, func_name)(regex=regex, **kwargs)
        except Exception as e:
            raise type(e)(f"failed to execute: {field} ; {expr}") from e

    return saqc



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import pandas as pd
import dios

from saqc.core import initFlagsLike, Flags
from saqc.constants import BAD, UNFLAGGED
from saqc.funcs.resampling import linear, interpolate, shift, concatFlags, resample
from saqc.funcs.tools import copyField, dropField
from tests.common import checkDataFlagsInvariants


@pytest.fixture
def data():
    index = pd.date_range(
        start="1.1.2011 00:00:00", end="1.1.2011 01:00:00", freq="15min"
    )
    index = index.insert(2, pd.Timestamp(2011, 1, 1, 0, 29, 0))
    index = index.insert(2, pd.Timestamp(2011, 1, 1, 0, 28, 0))
    index = index.insert(5, pd.Timestamp(2011, 1, 1, 0, 32, 0))
    index = index.insert(5, pd.Timestamp(2011, 1, 1, 0, 31, 0))
    index = index.insert(0, pd.Timestamp(2010, 12, 31, 23, 57, 0))
    index = index.drop(pd.Timestamp("2011-01-01 00:30:00"))
    dat = pd.Series(np.linspace(-50, 50, index.size), index=index, name="data")
    # good to have some nan
    dat[-3] = np.nan
    data = dios.DictOfSeries(dat)
    return data


@pytest.mark.parametrize(
    "func, kws",
    [
        ("linear", dict()),
        ("shift", dict(method="nshift")),
        ("interpolate", dict(method="spline")),
        ("resample", dict(func=np.nansum, method="nagg")),
    ],
)
def test_wrapper(data, func, kws):
    field = "data"
    freq = "15T"
    flags = initFlagsLike(data)

    import saqc

    func = getattr(saqc.funcs, func)
    data, flags = func(data, field, flags, freq, **kws)

    # check minimal requirements
    checkDataFlagsInvariants(data, flags, field)
    assert data[field].index.inferred_freq == freq


@pytest.mark.parametrize("method", ["time", "polynomial"])
def test_gridInterpolation(data, method):
    freq = "15T"
    field = "data"
    data = data[field]
    data = (data * np.sin(data)).append(data.shift(1, "2h")).shift(1, "3s")
    data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)

    # we are just testing if the interpolation gets passed to the series without
    # causing an error:
    res = interpolate(
        data.copy(),
        field,
        flags.copy(),
        freq,
        method=method,
        downcast_interpolation=True,
    )

    if method == "polynomial":
        res = interpolate(
            data.copy(),
            field,
            flags.copy(),
            freq,
            order=2,
            method=method,
            downcast_interpolation=True,
        )
        res = interpolate(
            data.copy(),
            field,
            flags.copy(),
            freq,
            order=10,
            method=method,
            downcast_interpolation=True,
        )

    # check minimal requirements
    rdata, rflags = res
    checkDataFlagsInvariants(rdata, rflags, field, identical=False)
    assert rdata[field].index.inferred_freq == freq


@pytest.mark.parametrize(
    "func, kws",
    [
        ("linear", dict()),
        ("shift", dict(method="nshift")),
        ("interpolate", dict(method="spline")),
        ("aggregate", dict(value_func=np.nansum, method="nagg")),
    ],
)
def test_flagsSurviveReshaping(func, kws):
    """
    flagging -> reshaping -> test (flags also was reshaped correctly)
    """
    pass


def test_flagsSurviveInverseReshaping():
    """
    inverse reshaping -> flagging -> test (flags also was reshaped correctly)"""
    pass


def test_flagsSurviveBackprojection():
    """
    flagging -> reshaping -> inverse reshaping -> test (flags == original-flags)
    """
    pass


@pytest.mark.parametrize(
    "reshaper", ["nshift", "fshift", "bshift", "nagg", "bagg", "fagg", "interpolation"]
)
def test_harmSingleVarIntermediateFlagging(data, reshaper):
    flags = initFlagsLike(data)
    field = "data"
    freq = "15T"

    pre_data = data.copy()
    pre_flags = flags.copy()
    data, flags = copyField(data, field, flags, field + "_interpolated")
    data, flags = linear(data, field + "_interpolated", flags, freq=freq)
    checkDataFlagsInvariants(data, flags, field + "_interpolated", identical=True)
    assert data[field + "_interpolated"].index.inferred_freq == freq

    # flag something bad
    flags[data[field + "_interpolated"].index[3:4], field + "_interpolated"] = BAD
    data, flags = concatFlags(
        data, field + "_interpolated", flags, method="inverse_" + reshaper, target=field
    )
    data, flags = dropField(data, field + "_interpolated", flags)

    assert len(data[field]) == len(flags[field])
    assert data[field].equals(pre_data[field])
    assert flags[field].index.equals(pre_flags[field].index)

    if "agg" in reshaper:
        if reshaper == "nagg":
            start, end = 3, 7
        elif reshaper == "fagg":
            start, end = 3, 5
        elif reshaper == "bagg":
            start, end = 5, 7
        else:
            raise NotImplementedError("untested test case")

        assert all(flags[field].iloc[start:end] > UNFLAGGED)
        assert all(flags[field].iloc[:start] == UNFLAGGED)
        assert all(flags[field].iloc[end:] == UNFLAGGED)

    elif "shift" in reshaper:
        if reshaper == "nshift":
            exp = [False, False, False, False, True, False, False, False, False]
        elif reshaper == "fshift":
            exp = [False, False, False, False, True, False, False, False, False]
        elif reshaper == "bshift":
            exp = [False, False, False, False, False, True, False, False, False]
        else:
            raise NotImplementedError("untested test case")

        flagged = flags[field] > UNFLAGGED
        assert all(flagged == exp)

    elif reshaper == "interpolation":
        pytest.skip("no testcase for interpolation")

    else:
        raise NotImplementedError("untested test case")


@pytest.mark.parametrize(
    "params, expected",
    [
        (
            ("nagg", "15Min"),
            pd.Series(
                data=[-87.5, -25.0, 0.0, 37.5, 50.0],
                index=pd.date_range(
                    "2011-01-01 00:00:00", "2011-01-01 01:00:00", freq="15min"
                ),
            ),
        ),
        (
            ("nagg", "30Min"),
            pd.Series(
                data=[-87.5, -25.0, 87.5],
                index=pd.date_range(
                    "2011-01-01 00:00:00", "2011-01-01 01:00:00", freq="30min"
                ),
            ),
        ),
        (
            ("bagg", "15Min"),
            pd.Series(
                data=[-50.0, -37.5, -37.5, 12.5, 37.5, 50.0],
                index=pd.date_range(
                    "2010-12-31 23:45:00", "2011-01-01 01:00:00", freq="15min"
                ),
            ),
        ),
        (
            ("bagg", "30Min"),
            pd.Series(
                data=[-50.0, -75.0, 50.0, 50.0],
                index=pd.date_range(
                    "2010-12-31 23:30:00", "2011-01-01 01:00:00", freq="30min"
                ),
            ),
        ),
    ],
)
def test_harmSingleVarInterpolationAgg(data, params, expected):
    flags = initFlagsLike(data)
    field = "data"
    h_field = "data_harm"

    pre_data = data.copy()
    pre_flaggger = flags.copy()
    method, freq = params

    data_harm, flags_harm = copyField(data, "data", flags, "data_harm")
    data_harm, flags_harm = resample(
        data_harm, h_field, flags_harm, freq, func=np.sum, method=method
    )
    checkDataFlagsInvariants(data_harm, flags_harm, h_field, identical=True)
    assert data_harm[h_field].index.freq == pd.Timedelta(freq)
    assert data_harm[h_field].equals(expected)

    data_deharm, flags_deharm = concatFlags(
        data_harm, h_field, flags_harm, target=field, method="inverse_" + method
    )
    data_deharm, flags_deharm = dropField(data_deharm, h_field, flags_deharm)
    checkDataFlagsInvariants(data_deharm, flags_deharm, field, identical=True)
    assert data_deharm[field].equals(pre_data[field])
    assert flags_deharm[field].equals(pre_flaggger[field])


@pytest.mark.parametrize(
    "params, expected",
    [
        (
            ("bshift", "15Min"),
            pd.Series(
                data=[-50.0, -37.5, -25.0, 12.5, 37.5, 50.0],
                index=pd.date_range(
                    "2010-12-31 23:45:00", "2011-01-01 01:00:00", freq="15Min"
                ),
            ),
        ),
        (
            ("fshift", "15Min"),
            pd.Series(
                data=[np.nan, -37.5, -25.0, 0.0, 37.5, 50.0],
                index=pd.date_range(
                    "2010-12-31 23:45:00", "2011-01-01 01:00:00", freq="15Min"
                ),
            ),
        ),
        (
            ("nshift", "15min"),
            pd.Series(
                data=[np.nan, -37.5, -25.0, 12.5, 37.5, 50.0],
                index=pd.date_range(
                    "2010-12-31 23:45:00", "2011-01-01 01:00:00", freq="15Min"
                ),
            ),
        ),
        (
            ("bshift", "30Min"),
            pd.Series(
                data=[-50.0, -37.5, 12.5, 50.0],
                index=pd.date_range(
                    "2010-12-31 23:30:00", "2011-01-01 01:00:00", freq="30Min"
                ),
            ),
        ),
        (
            ("fshift", "30Min"),
            pd.Series(
                data=[np.nan, -37.5, 0.0, 50.0],
                index=pd.date_range(
                    "2010-12-31 23:30:00", "2011-01-01 01:00:00", freq="30Min"
                ),
            ),
        ),
        (
            ("nshift", "30min"),
            pd.Series(
                data=[np.nan, -37.5, 12.5, 50.0],
                index=pd.date_range(
                    "2010-12-31 23:30:00", "2011-01-01 01:00:00", freq="30Min"
                ),
            ),
        ),
    ],
)
def test_harmSingleVarInterpolationShift(data, params, expected):
    flags = initFlagsLike(data)
    field = "data"
    h_field = "data_harm"
    pre_data = data.copy()
    pre_flags = flags.copy()
    method, freq = params

    data_harm, flags_harm = copyField(data, "data", flags, "data_harm")
    data_harm, flags_harm = shift(data_harm, h_field, flags_harm, freq, method=method)
    assert data_harm[h_field].equals(expected)
    checkDataFlagsInvariants(data_harm, flags_harm, field, identical=True)

    data_deharm, flags_deharm = concatFlags(
        data_harm, h_field, flags_harm, target=field, method="inverse_" + method
    )
    checkDataFlagsInvariants(data_deharm, flags_deharm, field, identical=True)

    data_deharm, flags_deharm = dropField(data_deharm, h_field, flags_deharm)
    assert data_deharm[field].equals(pre_data[field])
    assert flags_deharm[field].equals(pre_flags[field])
